/*
** Module   : SecurityManager.java
** Abstract : security kernel
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description------------------------------
** 001 NVS 20050208   @19692 Created initial implementation.
** 002 NVS 20050412   @20687 Variables and functions are supported through
**                           the library interface. Library methods have 
**                           easy access to Rights instance.
** 003 NVS 20050414   @20714 Debug output methods are made public to be
**                           accessible from resource plugins.
** 004 NVS 20050421   @20871 Passing a generic Object between compute and
**                           evaluate calls and the resource plugin's
**                           library of exported variables and methods
**                           instead of Rights. This allows arbitrary
**                           plugin's object to serve as a convenient
**                           data linkage.
** 005 NVS 20050425   @20872 SecurityManager no longer mediates the batch
**                           editing sessions. Applications call directory
**                           service directly to do directory editing.
**                           Directory service uses these notification
**                           methods of SecurityManager to signal batch
**                           related events: openBatch(), isEditing() and
**                           closeBatch().
** 006 NVS 20050427   @20913 Enhanced to provide unlimited storage for any
**                           kind of tokens (objects) that can be associa-
**                           ted with this context. The need is to get rid
**                           of all sorts of per thread state information
**                           and replace it with per context one. This
**                           overcomes the problem of arbitrary chosen
**                           threads from the thread pool to serve eleme-
**                           ntary requests which have to share some state
**                           for the lifespan of the session.
** 007 NVS 20050429   @20941 Fixed bugs in access control code for open,
**                           closeBatch and isEditing methods.
** 008 NVS 20050504   @21066 Implemented finalization of security contexts
**                           Contexts are assigned to sessions upon crea-
**                           tion. terminateSession() unassigns context 
**                           from session and deletes session. assign() 
**                           and unassign() calls are balanced. The unas-
**                           sign() call which zeros the use counter is 
**                           responsible for calling cleanup() method for 
**                           the security context.
** 009 NVS 20050505   @21107 Some major rework of the initial security
**                           context handling and the way the generations 
**                           of the security caches coexist to fix the bug
**                           documented in ##21100-21105.
** 010 NVS 20050524   @21273 This class now implements the Scope interface
**                           which is required by the new Expression Engi-
**                           ne. The scopes are nested so that plugins 
**                           have the Security Manager's scope as their
**                           parent. Also, instead of separate compute()
**                           and evaluate() methods a generic evaluate()
**                           method is implemented which returns Object.
** 011 NVS 20050725   @21816 Implemented the getUserId() method required
**                           to support the Progress USERID builtin func-
**                           tion.
** 012 NVS 20050726   @21828 Added methods to retrieve and set arbitrary
**                           account extension data of types int, boolean,
**                           String and byte[].
**                           method is implemented which returns Object.
** 013 SIY 20050726   @21830 Changed DirectoryService method name for
**                           restricted use method call closeBatch from
**                           closeBatch to closeBatchInt.
** 014 NVS 20050727   @21868 Initial security context of security cache is
**                           now permanently bound to DirectoryService
**                           upon creation. The call to securityIsUp 
**                           method relocated from SecurityCache constru-
**                           ctor to SecurityManager constructor as this 
**                           is one time thing.
** 015 NVS 20050812   @22089 Bug fix: anonymous processes could not login.
**                           The problem was in authenticateLocal() method
**                           trying to pass the anonymous process ID as a
**                           user ID to createSecurityContext() call.
** 016 NVS 20050831   @22446 Directory service is instantiated outside of
**                           and before the security manager. This allows
**                           directory based initialization of the server
**                           application.
** 017 NVS 20050912   @22692 Implemented authentication errors recovery. 
**                           For those cases where the user input is 
**                           required (ID + PW, Certificate + ID + PW, 
**                           Custom), authentication protocol allows for a
**                           configurable number of retries (unlimited 
**                           retries and no retries are also possible).
** 018 NVS 20050914   @22710 Added clientFinalize() method to comply with
**                           the changed Authenticator interface.
** 019 NVS 20051003   @22926 Fixed a small bug in custom authentication
**                           hook invocation failure handling.
** 020 ECF 20060120   @24006 Added variants of {add|remove|has|get}Token
**                           which are package private and use Object keys
**                           instead of Strings. The public, String-key
**                           variants of these methods are now deprecated.
**                           They must be removed once client code
**                           dependencies have been resolved.
** 021 GES 20060122   @24012 Added getAccountIds() method to provide a
**                           list of all valid subject IDs for the current
**                           context.
** 022 ECF 20060123   @24036 Added session and thread ID constructs. A
**                           unique session ID is assigned to each context
**                           as it is created. A unique thread ID is
**                           mapped to a thread at the time it is assigned
**                           a context (an unmapped when the context is
**                           unassigned). Both are accessed via read-only
**                           getter methods.
** 023 NVS 20060217   @24614 Server's certificate validation is performed
**                           only if explicitly or implicitly requested.
**                           It is requested implicitly if client runs
**                           with either keystore or truststore or both.
**                           If neither is specified, the validation can
**                           be requested explicitly using "security",
**                           "certificate", "validate" config item set to
**                           "true".
** 024 GES 20060305   @24897 Reworked logging to use J2SE instead of
**                           stdio.
** 025 GES 20060313   @25014 Fixed logging init to set the default level.
** 026 NVS 20060313   @25155 Fixed synchronization bug in an inner class.
**                           Removed some debugging output.
** 027 NVS 20060328   @25243 Added two new getExtDate(...) methods.
** 028 NVS 20060330   @25285 Added new methods: isUserDefined() and 
**                           getAllUsers().
** 029 NVS 20060331   @25406 Added new methods: getIdByExtInteger() and
**                           getIdByExtString().
** 030 NVS 20060412   @25503 Added new methods:getAllSubjects() and 
**                           getIdByOrdinal(). Protected all recently
**                           introduced methods with the system/accounts
**                           and system/extensions resource access rights
**                           check.
** 031 NVS 20060414   @25575 Added new method setPassword() that supports
**                           dynamic password change. Added new method
**                           changePassword() that initiates the password
**                           change procedure from application.
** 032 NVS 20060418   @25584 Password input hook instantiation had to be
**                           delayed till the use time.
** 033 NVS 20060426   @25729 Fixed thread ID assignment for the purpose of
**                           describing security contexts.
** 034 NVS 20060426   @25744 Fixed password change implementation to try
**                           adding password if there is no one yet. Fixed
**                           password age calculation.
** 035 ECF 20060529   @26596 Replaced WeakHashMap with HashMap for thread
**                           IDs. Changed default logging output to STAT.
** 036 ECF 20060614   @27245 Optimized auditing. Only compose audit
**                           entries if auditing is enabled.
** 037 NVS 20060426   @27378 authenticateClient now is called from gearUp
**                           method of ClientBootstrap.
** 038 GES 20060728   @28212 Rework of the checkCaller() interfaces and
**                           backing methods to use common code and to
**                           be *significantly* faster. The largest
**                           performance enhancement comes from the
**                           array form where previous usage got a new
**                           copy of the same call stack for every entry 
**                           in the array that needed to be checked.
** 039 GES 20061002   @30110 Reworked client authentication processing to
**                           pass a return code which specifies a more
**                           precise reason for any failure.  This allows
**                           the client auth hook to implement different
**                           logic. As part of this change, much of the
**                           code was reworked, refactored and cleaned
**                           up.  In addition, the requirement that there
**                           be a server-side custom auth plugin is eased
**                           (depending on the way it is registered) in
**                           the case that the registration is done of
**                           a base name that then must have the "Client"
**                           added to it.  If one explicitly registers a
**                           full class name, then this will be used on
**                           the server too.
** 040 GES 20061010   @30314 Honor a configuration value for debug level
**                           which optionally may be present in the 
**                           directory.  If present, this overrides the
**                           default debug level which is hard coded in
**                           this class.
** 041 GES 20061215   @31708 Changes in bootstrap config interface.
** 042 GES 20070111   @31800 Match changes in the Authenticator interface.
**                           Handled some cleanup and refactoring 
**                           including the elimination of AuthUIHelper and
**                           the simplification of external interfaces.
**                           Matched caller name changes and exposed
**                           stack checking methods externally.
** 043 GES 20070115   @31839 Changed authorized callers for context mgmt.
** 044 GES 20070118   @31878 Added some TODOs, documentation and fixed
**                           a potential null pointer exception.
** 045 NVS 20070316   @32414 Added closeBatch() method with controlled
**                           security cache refresh.
** 046 NVS 20070411   @32933 Security threads are created as part of a
**                           thread group named "security". These are
**                           security cache refresh thread ("refresh") and
**                           password change thread ("passwordchange").
**                           Transient thread IDs are now deleted from the
**                           thread ID map when they drop the initial
**                           security context.
** 047 NVS 20070412   @32967 Reworked security context cleanup. cleanup()
**                           method of SecurityContext, which is made
**                           package private, is called from 
**                           popContextWorker() method when the context is
**                           going out of use. The context remains 
**                           accessible for the current thread for the
**                           lifespan of the cleanup() method call.
** 048 GES 20071011   @35428 Added safety code and cleaned up misc code
**                           formatting/docs.
** 049 ECF 20071112   @35902 Changed many methods to accommodate net
**                           package refactoring. Most restricted use
**                           locations changed as a result. Added method
**                           getEffectiveContext() (restricted) to query
**                           effective context from net package. Removed
**                           methods deprecated in #020 (@24006), as all
**                           dependencies on these methods have now been
**                           removed from use.
** 050 ECF 20071127   @36046 Fixed a minor defect in authenticateLocal().
**                           Replaced & operator where && was required.
** 051 NVS 20080513   @38274 Fixed context cleanup when it happens in
**                           the reader thread through terminateSession()
**                           method. Fixed logging for going context.
** 052 ECF 20081006   @40039 Added variant of debug(). Accepts an
**                           Exception in order to output a stack trace.
** 053 GES 20081024   @40341 Added support for sessions to be established
**                           on insecure sockets (non-SSL).
** 054 NVS 20090130   @41245 Added support for programmatic authentication
**                           type which is intended for use with admin
**                           applet etc. The idea is that the applet needs
**                           ID+PW authentication no matter what type is
**                           set on the server for interactive users (this
**                           may be a custom authentication plugin). This
**                           authentication type is less secure per se,
**                           AND, given the fact that it will be used to
**                           administer the system, the implementation has
**                           to be fortified somehow A.S.A.P. (since the
**                           applet that uses this authentication gets
**                           downloaded from the web server part of the
**                           P2J server, the applet can get a token with a
**                           limited lifespan that is also valid only if
**                           returned from the same IP address). 
** 055 JJC 20090208   @41406 Added FILE_SEPARATOR = "\\" + File.separator,
**                           which works on Windows and Linux.
**                           Changed logging so that the console handler
**                           is turned off if not on the server. This
**                           avoids permissions issue in the admin applet.
** 056 NVS 20090312   @41523 Added adminGetServerName() method.
** 057 NVS 20090318   @41629 Fixed a latent bug in SecurityCacheRefresh
**                           inner class. Added adminGetCacheSerial() and
**                           adminRefresh() methods.
** 058 NVS 20090323   @41668 Client side logic now allows the bootstrap
**                           configuration of the same type to replace the
**                           cached one.
** 059 NVS 20090326   @41683 Some changes due to creation of SecurityAdmin
**                           class.
** 060 NVS 20090415   @41801 Cache generation messages can be routed to
**                           admin client, server console or both ways.
** 061 NVS 20090417   @41809 Fixed a NPE in getIdByExtInteger() method and
**                           prevented a similar NPE in getIdByExtString().  
** 062 SIY 20090515   @42189 Added one more legal caller for  
**                           setInitialSecurityContext().
** 063 NVS 20090521   @42421 Now using an external implementation of the 
**                           password hashing algorithm.
** 064 NVS 20090528   @42481 Custom account extension plugins are read 
**                           from the directory and made available through
**                           getCustomServerExt(), getCustomClientExt().
** 065 NVS 20090602   @42549 Custom account extension plugin jar file name
**                           is read from the directory and made available
**                           through getCustomClientJar() method.
** 066 NVS 20090603   @42597 Supporting account enablement status.
** 067 NVS 20090604   @42598 Supporting explicit password protected status.
** 068 NVS 20090615   @42697 Removed #065 due to a utilization of a simpler
**                           approach.
** 069 GES 20090617   @42762 Added getSessionReport().
** 070 GES 20090619   @42784 Added killSession(). Fixed inherent problems with
**                           authenticateClientWorker() which temporarily
**                           switched bootstrap config instances for the
**                           singleton security manager instance WITHOUT any
**                           protection and WITHOUT any assurance that other
**                           SM code wouldn't access the (invalid client cfg)
**                           during the same period of time. This would lead to
**                           unexpected failures that couldn't be easily
**                           reproduced. This stems from the dual usage of
**                           that worker for server-to-server connections. Also
**                           in that case, no session object is kept in the
**                           list of sessions since normal termination rules
**                           do not apply.
** 071 NVS 20090624   @42843 Initial startup may need to write to the directory
**                           for converting ACLs. Made it possible through
**                           checks of the initialized status.
** 072 ECF 20090813   @43614 Added isGroupDefined(). Refactored worker code
**                           from isUserDefined() into getAccount() method.
** 073 GES 20090816   @43640 Allow terminateSession() from Queue.stop().
** 074 NVS 20090816   @43651 The custom auth plugin can tell the server it wants
**                           to exit silently by returning an empty array of
**                           authentication data.
** 075 CA  20090824   @43721 The "Error getting AUTHDATA" message in 
**                           authenticateLocal is no longer logged as SEVERE.
** 076 CA  20090902   @43822 More log-level changes in authenticateLocal - 
**                           SocketException's are caught and logged using 
**                           dTrace().
** 077 NVS 20090909   @43849 Fixed retry loop in authenticateClient(). The loop
**                           is interrupted in AUTH_MODE_IDPW if the source of
**                           the input is the bootstrap config, since the result
**                           won't change.
** 078 NVS 20090918   @43954 Made hasToken() and getToken() methods public.
** 079 NVS 20090925   @44036 Password change is better synchronized with the
**                           security cache refresh. The changed password is
**                           always stored it the most recent secury cache
**                           generation.
** 080 NVS 20091007   @44109 adminAccess() queries the AdminResource plugin and
**                           falls back to the SystemResource if not installed.
** 081 GES 20091103   @44306 Added logging to client authentication and allowed
**                           changing the logging level on a client system.
** 082 CA  20091026   @44197 Added methods to build a consolidated set of ACLs. 
**                           This implies adding new permission search methods,
**                           to determine the rights for a subject other than 
**                           the authenticated one.
** 083 CA  20091119   @44432 Added SessionToken implementation - an unique
**                           object is kept for each security context, which
**                           will hold the session ID and the subject.
** 084 GES 20100519   @44859 Virtual sessions now have a string as their session
**                           ID object instead of a NetSocket. This breaks
**                           session kill and session listing for the admin
**                           interface.  This code is now protected.
** 085 SVL 20100604   @44901 Added isProcessDefined function.
** 086 SVL 20100706          Fixed authentication for users using only a
**                           certificate: it does not fall to custom 
**                           authentication anymore.
** 087 SVL 20100709          Do not allow a user to login using user ID +
**                           password if he has "certificate only" auth type.
** 088 SVL 20100713          Do don't override default auth mode by user's 
**                           auth mode AUTH_MODE_NONE.
** 089 SVL 20100916          User auth mode and auth plugin are preserved.
**                           Client can send target user id at the beginning
**                           of the authentication process.
** 090 SVL 20100923          Changed to conform the new Authenticator API. 
** 091 CA  20101214          Solved memory leak on client side, when multiple
**                           sessions are being created during the same run
**                           (no security.Session needs to be created on 
**                           client side, as there is no security context).
** 092 CA  20110105          Fixed concurrency issues related to ACL searches.
**                           After authentication, send the authentication 
**                           details (process and/or user ID) back to the
**                           requester side.
** 093 CA  20110114          Fixed a NPE in rights searching (caused by H092).
** 094 CA  20110131          Changes to allow the direct sessions to authenti-
**                           cate using different credentials. Also, this way
**                           sessions can be created to different router nodes.
** 095 LMR 20101215          Changed authenticateClientWorker() to default
**                           net/connection/secure to false.
** 096 GES 20111003          Changed obtainCustomClientHook() to try to load
**                           a hook class from the classpath if the byte[]
**                           cannot be loaded (this happens when the client is
**                           running in an applet).
** 097 GES 20111028          Added authenticateSingle() and terminateSingle()
**                           which provide a mechanism to login and create an
**                           interactive client session where the client code
**                           runs in the server process. Normally, the security
**                           authentication and session termination is fully
**                           integrated and driven by the setup of a network
**                           session. In the local mode, there is no network
**                           session, but the security context management must
**                           still be done.
** 098 GES 20111110          Modifications to provide the ability to have
**                           unique initial contexts (that are separate from
**                           the standard server's initial context). These are
**                           used when the client is run in the server JVM and
**                           it is important to enable multiple clients in the
**                           same JVM (since they must not co-mingle their
**                           contexts or context-local data). It is also used
**                           to make sure no client state is mingled with
**                           the standard server thread's initial context
**                           (which is shared for all server threads).
** 099 CA  20111209          Added method to list active sessions which have
**                           invoked APIs in the given jar or interface.
**                           checkCallerAbort now allows to check functions
**                           of SecurityManager.
** 100 CA  20120727          The context cleanup code must be executed by only
**                           one thread.  If other threads reach the cleanup
**                           code for the same context, it will no-op when
**                           another thread has started the cleanup for the
**                           same context.
** 101 ECF 20130321          Enable authenticateRemote to work if only a process ID is provided.
** 102 CA  20130529          Changes for the appserver implementation.
** 103 CA  20130628          Added possibility to reset a context (needed by appservers running in
**                           State-reset operating mode).
** 104 CA  20130704          Added API to get the accounts for a specified appserver.
** 105 CA  20131122          Protected security "cache" field access against possible concurrency 
**                           problems.  During authentication: if a newer security cache 
**                           generation is referenced, do not allow authentication; always use
**                           the security cache generation saved at the begining of the auth
**                           process, throughout the authentication.
*/

package com.goldencode.p2j.security;

import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import java.util.*;
import java.util.logging.*;
import javax.net.ssl.*;
import com.goldencode.expr.*;
import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.util.*;

/**
 * This class is the core of P2J security management.
 * <p>
 * Details on client authentication protocol:
 * <br>
 * Once the TLS connection is established and certificates are exchanged and
 * verified:
 * <ul>
 *   <li>if the connection is anonymous (client didn't send certificate):
 *     <ul>
 *       <li>the client sends the AUTHTYPE packet, where the content can be 
 *           either AUTH_REQ_PROCESS or AUTH_REQ_USER; if AUTHTYPE is
 *           AUTH_REQ_USER then id of the target user is send (it may be empty
 *           if it is not know at the beginning of authentication process).
 *       <li>for AUTH_REQ_PROCESS, the server immediately sets AUTH_MODE_REQ
 *           to either SUCCESS (if anonymous connections are allowed, or
 *           FAILURE otherwise;
 *       <li>the case of AUTH_REQ_USER processing continues with the next step
 *     </ul>
 *   <li>the server sends the AUTHMODEREQ packet;
 *   <li>the client reads the AUTHMODEREQ packet (with an authentication mode
 *       and a disposition/action code) and checks the disposition;
 *       if the disposition is either AUTH_ACTION_ABORT (on fatal error) or
 *       AUTH_ACTION_DONE (on success due to certificate based or anonymous 
 *       authenticaion), the procedure is over; otherwise it continues with
 *       the next step;
 *   <li>if the mode is AUTH_MODE_CUSTOM, the server should have sent and
 *       the client reads AUTHCUSTOM packet which delivers the class name and
 *       parameters of the custom authentication hook;  this hook class is
 *       run (which can display any UI and interact with the user OR it may
 *       access any device/facility such as a smart card);  the result is a
 *       userid and password
 *   <li>if the mode is AUTH_MODE_IDPW or AUTH_MODE_X509_IDPW, then a generic 
 *       client side processing is performed which reads the userid and 
 *       password
 *   <li>either custom or standard modes produce a byte array of 
 *       authentication data which are sent in AUTHDATA packet back to server;
 *   <li>server responds with AUTHRESULT packet (includes a return code and
 *       a disposition/action code) which reports either AUTH_ACTION_ABORT
 *       (on fatal failure), AUTH_ACTION_RETRY (retryable failure) or
 *       AUTH_ACTION_DONE (on success)
 *   <li>on the client, any action of AUTH_ACTION_RETRY will cause the
 *       processing to repeat starting with the AUTHMODEREQ packet; note that
 *       the return code sent in the AUTHRESULT packet will be provided to 
 *       the authentication plugin at that time;
 *   <li>on the client, any action of AUTH_ACTION_ABORT causes the processing 
 *       to halt and any action of AUTH_ACTION_DONE authentication has 
 *       successfully completed, either way the processing is over;
 * </ul>
 * The AUTHDATA packet for the standard modes is made of a <code>String</code>
 * userID followed by <code>char[]</code> password. UserID is serialized using
 * <code>writeUTF()</code> method. Password is serialized using 
 * <code>writeShort()</code> encoding the array length, plus 
 * <code>writeChar</code> for every character in the array.
 * <p>
 * A unique session ID is assigned to each security context upon creation.
 * Whenever a thread is assigned to a context, it is mapped to a new ID.
 * While the thread IDs (which are integers) could theoretically wrap, such
 * that a thread associated with a very long running context could be assigned
 * an ID already in use, the combination of session ID and thread ID is
 * guaranteed always to be unique, and it is really only this combination
 * which is meaningful for practical use.
 */
public final class SecurityManager
implements Authenticator,
           Scope
{
   /** file separator that works on both Windows and Linux. */
   public static final String  FILE_SEPARATOR = "\\" + File.separator;
        
   /** Logging level NONE. */
   public static final Level NONE = Level.OFF;
   
   /** Logging level ERROR. */
   public static final Level ERROR = Level.SEVERE;
   
   /** Logging level WARNING. */
   public static final Level WARNING = Level.WARNING;
   
   /** Logging level STAT. */
   public static final Level STAT = Level.INFO;
   
   /** Logging level LIST. */
   public static final Level LIST = Level.FINE;
   
   /** Logging level TRACE. */
   public static final Level TRACE = Level.FINER;
   
   /** Logging level DATA. */
   public static final Level DATA = Level.FINEST;
   
   /** Reference to the singleton instance. */
   private static SecurityManager securityManager = null;

   /** Thread group for all security threads. */
   private static ThreadGroup secThreads = new ThreadGroup(
      Thread.currentThread().getThreadGroup(), "security");

   /** The bootstrap configuration. */
   private BootstrapConfig cfg = null;

   /** Caches context data. */
   private SecurityCache cache = null;
   
   /** Last used serial number for security cache creation. */
   private int cacheSerial = 0;
   
   /** Provides network socket security services. */
   private TransportSecurity tranSec = null;
   
   /** List of sessions. */
   private List listOfSess = null;
   
   /** Anonymous log instance. */
   private Logger LOG = null;

   /** Debug level currently in effect. */
   private Level debugLevel = STAT;

   /** Debug level requested. */
   private Level debugLevelRequest = null;
   
   /** Next available, unique user session ID. */
   private int nextUserSessionID = 0;
   
   /** Next available, unique thread ID. */
   private int nextThreadID = 0;
   
   /** Map of threads to their current IDs. */
   private Map threadIDs = new HashMap();
   
   /** Flag to detect if we are initialized or not. */
   private boolean init = false;

   /** Object for serializing password change requests. */
   private Object pwchSync = new Object();
   
   /** Map containing flags which inform other threads that cleanup for a 
    * certain context is in progress, so they should skip it. */
   private Map<SecurityContext, Boolean> cleanupLocks = 
      new HashMap<SecurityContext, Boolean>();

   /**
    * Private constructor to enable this as a singleton class. Instantiation
    * occurs via {@link #createInstance}.
    *
    * @param  bc               
    *         an instance of <code>BootstrapConfig</code> class used to get 
    *         the security related configuration information  
    *
    * @throws ConfigurationException
    *         may be thrown in SecurityCache
    * @throws NoSuchMethodException
    *         may be thrown in SecurityCache
    */
   private SecurityManager(BootstrapConfig bc)
   throws ConfigurationException,
          NoSuchMethodException
   {
      securityManager = this;

      // keep a copy of the bootstrap configuration
      cfg = bc;

      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      // override debug level from directory
      Level lvl = readDebugLevel(ds);
      
      if (lvl != null)
         debugLevel = lvl;
      
      // initialize logging (get an anonymous logger)
      LOG = LogHelper.getSecureLogger(bc.isServer());
      LOG.setLevel(debugLevel);

      // instantiate the security cache
      SecurityCache newCache = new SecurityCache(bc, cacheSerial);
      Audit audit = newCache.getAudit();

      setCache(newCache);

      // initialize list of sessions
      listOfSess = new LinkedList();

      if (!bc.isServer())
         return;

      ds.securityIsUp();

      try
      {
         audit.start();
      }
      catch (IOException exc)
      {
         // The newly defined audit log file can't be activated and
         // the old audit log's gone.
         dErr("*** Audit log failure *** " + exc);
      }

      SecurityContext ctx = newCache.getInitialSecurityContext();
      logNewContext(newCache, ctx);

      try
      {
         setInitialSecurityContext();
      }
      catch (RestrictedUseException exc)
      {
         throw new ConfigurationException("setInitialSecurityContext()", exc);
      }
      
      init = true;
   }
   
   /**
    * Instantiates SecurityManager as a singleton.
    *
    * @param  bc
    *         an instance of <code>BootstrapConfig</code> class used to get 
    *         the security related configuration information  
    *
    * @return instance of <code>SecurityManager</code>
    *
    * @throws ConfigurationException
    *         may be thrown in SecurityManager constructor
    * @throws NoSuchMethodException
    *         may be thrown in SecurityCache
    */
   public static synchronized
   SecurityManager createInstance(BootstrapConfig bc)
   throws ConfigurationException,
          NoSuchMethodException
   {
      if (securityManager == null)
      {
         securityManager = new SecurityManager(bc);
      }
      else
      {
         // if on the client and with the same client type config
         if (!securityManager.cfg.isServer() && !bc.isServer())
         {
            // replace the config with the new one
            securityManager.cfg = bc;
         }
      }

      return securityManager;
   }

   /**
    * Returns the reference to the instance of SecurityManager.
    *
    * @return instance of <code>SecurityManager</code>
    */
   public static synchronized SecurityManager getInstance()
   {
      return securityManager;
   }
   
   /**
    * Returns a flag indicating the readiness of the security manager to
    * properly operate.
    *
    * @return   <code>true</code> if the security manager is ready to accept
    *           requests.
    */
   public static synchronized boolean initialized()
   {
      return (securityManager != null && securityManager.init);
   }

   /**
    * Gets custom server extension plugin name. 
    *
    * @return  custom server extension plugin name
    */
   public String getCustomServerExt()
   {
      return getCache().getCustomServerExt();
   }
   
   /**
    * Gets custom client extension plugin name. 
    *
    * @return  custom client extension plugin name
    */
   public String getCustomClientExt()
   {
      return getCache().getCustomClientExt();
   }
   
   /**
    * Package a user ID and password into a single entity for transmission.
    *
    * @param    userid
    *           The subject ID.
    * @param    password
    *           The password.
    *
    * @return   The object containing the packaged values.
    */
   public static byte[] packageIdPassword(String userid, String password)
   {
      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      DataOutputStream      dos = new DataOutputStream(bos);
      
      try
      {
         dos.writeUTF(userid);
         dos.writeUTF(password);
      }
      
      catch (IOException ioe)
      {
         return new byte[0];
      }
      
      return bos.toByteArray();      
   }
   
   /**
    * Checks the call stack to see if the caller is any of the methods
    * specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from both the <code>SecurityManager</code> and the given
    * <code>ignore</code> class name is the entry checked.
    * <p>
    * A function of <code>SecurityManager</code> can be specified as the
    * allowed function. In this case it should be the closest entry to the
    * top except <code>ignore</code> class functions and other functions of
    * <code>SecurityManager</code>.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A list of fully qualified (package.class.method) names from
    *           which the call is allowed.
    * @param    ignore
    *           A class name of entries (other than that of the
    *           <code>SecurityManager</code>) which should be ignored at
    *           the top of the stack.  May be <code>null</code>.
    *
    * @throws   RestrictedUseException
    *           If the check fails.
    */
   public static void checkCallerAbort(String[] allowed, String ignore)
   throws RestrictedUseException
   {
      if (!checkCallerWorker(allowed, null, ignore))
      {
         throw new RestrictedUseException("Unauthorized caller!");
      }
   }

   /**
    * Checks the call stack to see if the caller is any of the methods
    * specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from the <code>SecurityManager</code> is the entry checked.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A list of fully qualified (package.class.method) names from
    *           which the call is allowed.
    *
    * @throws   RestrictedUseException
    *           If the check fails.
    */
   public static void checkCallerAbort(String[] allowed)
   throws RestrictedUseException
   {
      checkCallerAbort(allowed, null);
   }

   /**
    * Checks the call stack to see if the caller is the method specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from both the <code>SecurityManager</code> and the given
    * <code>ignore</code> class name is the entry checked.
    * <p>
    * A function of <code>SecurityManager</code> can be specified as the
    * allowed function. In this case it should be the closest entry to the
    * top except <code>ignore</code> class functions and other functions of
    * <code>SecurityManager</code>.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A fully qualified (package.class.method) name from which the
    *           call is allowed.
    * @param    ignore
    *           A class name of entries (other than that of the
    *           <code>SecurityManager</code>) which should be ignored at
    *           the top of the stack.  May be <code>null</code>.
    *
    * @throws   RestrictedUseException
    *           If the check fails.
    */
   public static void checkCallerAbort(String allowed, String ignore)
   throws RestrictedUseException
   {
      StringBuffer sb = new StringBuffer();
      
      if (!checkCallerWorker(new String[] {allowed}, sb, ignore))
      {
         throw new RestrictedUseException("This restricted use method " +
                                          "cannot be called from " +
                                          sb.toString());
      }
   }

   /**
    * Checks the call stack to see if the caller is the method specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from the <code>SecurityManager</code> is the entry checked.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A fully qualified (package.class.method) name from which the
    *           call is allowed.
    *
    * @throws   RestrictedUseException
    *           If the check fails.
    */
   public static void checkCallerAbort(String allowed)
   throws RestrictedUseException
   {
      checkCallerAbort(allowed, null);
   }

   /**
    * Checks the call stack to see if the caller is the method specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from both the <code>SecurityManager</code> and the given
    * <code>ignore</code> class name is the entry checked.
    * <p>
    * A function of <code>SecurityManager</code> can be specified as the
    * allowed function. In this case it should be the closest entry to the
    * top except <code>ignore</code> class functions and other functions of
    * <code>SecurityManager</code>.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A fully qualified (package.class.method) name from which the
    *           call is allowed.
    * @param    ignore
    *           A class name of entries (other than that of the
    *           <code>SecurityManager</code>) which should be ignored at
    *           the top of the stack.  May be <code>null</code>.
    *
    * @return   <code>true</code> if allowed.
    */
   public static boolean checkCaller(String allowed, String ignore)
   {
      return checkCallerWorker(new String[] {allowed}, null, ignore);
   }

   /**
    * Checks the call stack to see if the caller is the method specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from both the <code>SecurityManager</code> is the entry checked.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A fully qualified (package.class.method) name from which the
    *           call is allowed.
    *
    * @return   <code>true</code> if allowed.
    */
   public static boolean checkCaller(String allowed)
   {
      return checkCaller(allowed, null);
   }

   /**
    * Parses a X.509 certificate and returns the CN component of the subject's
    * distinctive name. 
    *
    * @param  x509
    *         certificate to parse 
    * @return commonName part of the subject's distinctive name. May be empty.
    * @throws CertificateParsingException
    *         if DN is malformed or no CN was found in DN
    */
   static String getSubjectCommonName(X509Certificate x509)
   throws CertificateParsingException
   {
      String name = x509.getSubjectDN().getName();
      String comp = null;
      String cnam = null;
      String cval = null;
      int i = 0;
      int j = 0;
      
      // components of the DN are comma+spaces separated
      i = name.indexOf(',');
      if (i == -1)
         comp = name;
      else
         comp = name.substring(0, i);
      
      while (comp.length() > 0)
      {
         j = comp.indexOf('=');
         if (j == -1)
            throw new CertificateParsingException("malformed subject DN");
         cnam = comp.substring(0, j).trim();
         cval = comp.substring(j + 1);

         // check for CN or COMMONNAME 
         if (cnam.compareToIgnoreCase("cn")         == 0 ||
             cnam.compareToIgnoreCase("commonName") == 0)
            return cval;

         if (i == -1)
            break;
         name = name.substring(i + 1);
         i = name.indexOf(',');
         if (i == -1)
            comp = name;
         else
            comp = name.substring(0, i);
      }
      
      throw new CertificateParsingException("malformed subject DN, no CN");
   }
   
   /**
    * Read the debug level configured in the directory as a string node with
    * the path "/security/config/debug-level", convert this debug level name
    * to a valid debug level and return it.
    *
    * @param    ds
    *           The directory from which to read.  May be <code>null</code>
    *           if there is no directory service available (e.g. on the
    *           client).
    *
    * @return   The configured debug level or <code>null</code> if no debug
    *           level (or no valid level) is configured in the directory. If
    *           there is no valid directory, <code>null</code> will be
    *           returned.
    */
   private static Level readDebugLevel(DirectoryService ds)
   {
      Level dbg = null;
      
      // the directory is a server-side only feature, it is null on the
      // client
      if (ds != null)
      {
         String lvl = ds.getNodeString("/security/config/debug-level",
                                       "value");
         
         if ("NONE".equalsIgnoreCase(lvl))
         {
            dbg = NONE;
         }
         else if ("ERROR".equalsIgnoreCase(lvl))
         {
            dbg = ERROR;
         }
         else if ("WARNING".equalsIgnoreCase(lvl))
         {
            dbg = WARNING;
         }
         else if ("STAT".equalsIgnoreCase(lvl))
         {
            dbg = STAT;
         }
         else if ("LIST".equalsIgnoreCase(lvl))
         {
            dbg = LIST;
         }
         else if ("TRACE".equalsIgnoreCase(lvl))
         {
            dbg = TRACE;
         }
         else if ("DATA".equalsIgnoreCase(lvl))
         {
            dbg = DATA;
         }
      }
      
      return dbg;
   }

   /**
    * Checks the call stack to see if the caller is the method specified.
    * <p>
    * The closest entry to the top of the stack where the class name differs 
    * from both the <code>SecurityManager</code> and the given
    * <code>ignore</code> class name is the entry checked.
    * <p>
    * A function of <code>SecurityManager</code> can be specified as the
    * allowed function. In this case it should be the closest entry to the
    * top except <code>ignore</code> class functions and other functions of
    * <code>SecurityManager</code>.
    * <p>
    * This arrangement allows for multiple levels of calls within this class
    * and the class which is checking the caller without affecting the check.
    *
    * @param    allowed
    *           A list of fully qualified (package.class.method) names from
    *           which the call is allowed.
    * @param    entry
    *           Optional user-supplied buffer so that on failure the invalid
    *           value will be available.  May be <code>null</code> if not
    *           needed by the caller.
    * @param    ignore
    *           A class name of entries (other than that of the
    *           <code>SecurityManager</code>) which should be ignored at
    *           the top of the stack.  May be <code>null</code>.
    *
    * @return   <code>true</code> if allowed.
    */
   private static boolean checkCallerWorker(String[]     allowed,
                                            StringBuffer entry,
                                            String       ignore)
   {
      if (allowed == null)
         return false;
      
      if (entry == null)
         entry = new StringBuffer();
      
      // obtain the stack trace (expensive!)
      Throwable thread = new Throwable();
      thread.fillInStackTrace();
      StackTraceElement[] stack = thread.getStackTrace();
      
      // the following loop starts at the 1 index position because the
      // 0 index pos will always be the current method... which doesn't
      // need checking <g>
      
      for (int i = 1; i < stack.length; i++)
      {
         String possibleClass = stack[i].getClassName();

         // check ignore class
         if (ignore != null && possibleClass.equals(ignore))
            continue;

         entry.setLength(0);
         entry.append(possibleClass).append(".");
         entry.append(stack[i].getMethodName());
         String possibleMethod = entry.toString();

         for (int j = 0; j < allowed.length; j++)
         {
            if (possibleMethod.equals(allowed[j]))
            {
              return true;
            }
         }

         if (possibleClass.equals(
               "com.goldencode.p2j.security.SecurityManager"))
         {
            // avoid internal calls
            continue;
         }

         // no match
         break;
      }

      return false;
   }

   /**
    * Parses a X.509 certificate and returns a list of alternative names of a
    * specified type. The list may be empty if no such names are found. 
    *
    * @param  x509
    *         certificate to parse 
    * @param  altNameType
    *         integer type of alternative name to look for 
    * @return <code>List</code> of found names as <code>String</code>s. May be
    *         empty.
    * @throws CertificateParsingException
    */
   private static List fetchAltNames(X509Certificate x509, int altNameType)
   throws CertificateParsingException
   {
      List names = new LinkedList();
      
      Collection col = x509.getSubjectAlternativeNames();
      if (col != null)
      {
         Iterator iter = col.iterator();
         List l = null;
         Integer iType = null;
         
         while (iter.hasNext())
         {
            // every item in the collection is a list with two elements
            l = (List)iter.next();
            // first element of the list is an Integer entry type
            iType = (Integer)l.get(0);
            // if type is 2, the second element of the list is a name String
            if (iType.intValue() == altNameType)
               names.add(l.get(1));
         }
      }
      
      return names;
   }

   /**
    * Compares the list of DNS names with the given host name. Leftmost * in
    * DNS names like *.tld matches any.name.tld. Comparison is case
    * insensitive.
    *
    * @param  host
    *         name to check 
    * @param  list
    *         list of names 
    * @return index of matching name in the list or -1 of no matches found
    */
   private static int hostMatch(String host, List list)
   {
      String name = null;
      String tld  = null;
      int i = 0;
      
      for (i = 0; i < list.size(); i++)
      {
         name = (String)list.get(i);

         if (name.startsWith("*."))
         {
            // match *.tld case
            tld = name.substring(1);
            if (host.toUpperCase().endsWith(tld.toUpperCase()))
               return i;
         }
         else
         {
            // match regular case
            if (host.compareToIgnoreCase(name) == 0)
               return i;
         }
      }

      return -1; 
   }


   /**
    * Get an instance of the in-process authentication class.
    *
    * @return   The instance of the custom client authentication hook.
    */
   private Authenticator obtainCustomAuthHook(String hookName)
   {
      Authenticator auth = null;
      
      try
      {
         String path = buildClassFileName(hookName);
         
         if (ClassLoader.getSystemResource(path) != null)
         {
            Class hook = Class.forName(hookName);
            
            if (hook != null)
            {
               auth = (Authenticator) hook.newInstance();
            }
         }
      }
      
      catch (Exception e)
      {
         // something is wrong with the class
         dErr("Error obtaining the authentication class. " + e);
      }
      
      return auth;
   }
   
   /**
    * Gets custom authentication class and parameters from the server and runs
    * the specified method.
    * <p>
    * The AUTHCUSTOM packet is two consecutive strings written using the
    * <code>writeUTF()</code> method. The first is the class name, followed
    * by the parameters. The no parameter case is handled as if it was an
    * empty string.
    *
    * @param    dis
    *           The input stream to read from. 
    *
    * @return   The instance of the custom client authentication hook.
    */
   private Authenticator obtainCustomClientHook(DataInput dis)
   {
      byte[] data = null;
      byte classData[] = null;
      int  classSize = 0;
      String hookName = null;
      
      try
      {
         // the packet is made of three pieces:
         // 1 - byte array, copy of the class to run
         //     formatted as int + bytes, where int == 0 means no class
         // 2 - fully qualified class name formatted as UTF
         // 3 - string of class parameters formatted as UTF (not read here)

         // read piece 1 - copy of the class
         classSize = dis.readInt();
         if (classSize > 0)
         {
            classData = new byte[classSize];
            dis.readFully(classData);
         }

         // read piece 2 - class name
         hookName = dis.readUTF();
      }
      catch (IOException io)
      {
         // something's wrong with the socket anyway
         dErr("Error downloading the authentication class. " + io);
         return null;
      }
      
      // create an instance of the class and call clientAuthHook 
      try
      {
         Class client = null;

         if (classData != null)
         {
            try
            {
               HookClassLoader hookLoader = new HookClassLoader(classData);
               client = hookLoader.loadClass(hookName, true);
            }
            catch (Exception exc)
            {
               dErr("HookClassLoader failed. " + exc);
               
               // fall back plan (check the classpath for a class of the same
               // name (this path is used for applets)
               client = Class.forName(hookName);
            }
         }
         else
         {
            client = Class.forName(hookName);
         }

         Authenticator instance = (Authenticator) client.newInstance();
         return instance;
      }
      catch (Exception e)
      {
         // something's wrong with the class
         dErr("Error loading the authentication class. " + e);
         return null;
      }
   }

   /**
    * Creates a valid filename using a fully qualified class name.
    *
    * @param    name
    *           A fully qualified class name (without the ".class" at the
    *           end). Note that any "." characters are converted into the
    *           system-specific file separators.
    *
    * @return   The filename.
    */
   private static String buildClassFileName(String name)
   {
      String       path = name.replaceAll("\\.", FILE_SEPARATOR); 
      StringBuffer sb   = new StringBuffer(path).append(".class");
      
      return sb.toString();
   }
   
   /**
    * Checks the validity (existence) of a class, first creating a filename
    * using a given fully qualified class name.
    *
    * @param    name
    *           A fully qualified class name (without the ".class" at the
    *           end). Note that any "." characters are converted into path
    *           separators.
    *
    * @return   <code>true</code> if the class file exists.
    */
   private static boolean classExists(String name)
   {
      String path = buildClassFileName(name);
      return (ClassLoader.getSystemResource(path) != null);
   }
   
   /**
    * Gets a byte array filled with the binary class data for the given
    * class.
    *
    * @param    name
    *           A fully qualified class name (without the ".class" at the
    *           end).
    *
    * @return   The array of bytes of the binary class file image.
    */
   private static byte[] getClassBytes(String name)
   {
      // compose resource name
      String path = buildClassFileName(name);
      InputStream in = ClassLoader.getSystemResourceAsStream(path);
      
      if (in == null)
         return null;

      // create the output stream for byte array
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      final int BUFFSIZE = 8192;
      byte[] buff = new byte[BUFFSIZE];
      int count = 0;

      // read the class file manually
      try
      {
         do
         {
            count = in.read(buff);
            if (count > 0)
               baos.write(buff, 0, count);
         } while (count > 0);
      }
      catch (IOException exc)
      {
         return null;
      }
      finally
      {
         try
         {
            in.close();
         }
         catch (IOException exc)
         {
         }
      }
      
      // return the class data
      return baos.toByteArray();
   }

   /**
    * Gets the server's account name.
    * This is an admin API support method.
    *
    * @param  scope
    *         <code>true</code> means using the latest security cache,
    *         otherwise the session security cache
    *
    * @return server's name or <code>null</code>
    */
   public String adminGetServerName(boolean scope)
   {
      if (!adminAccess())
      {
         return null;
      }
      
      SecurityCache sc = null;
      
      if (scope)
      {
         sc = getCache();
      }
      else
      {
         // Pick up the generation of security cache
         SecurityContextStack ctxs = SecurityContextStack.getContext();
         SecurityContext ctx = ctxs.getEffectiveContext();
         sc = ctx.getCache(); 
      }
      
      return sc.getServerAccount().getSubjectId();
   }

   /**
    * Gets the server's account name.
    * This is an admin API support method.
    *
    * @param  scope
    *         <code>true</code> means using the latest security cache,
    *         otherwise the session security cache
    *
    * @return security cache generation serial number
    */
   public int adminGetCacheSerial(boolean scope)
   {
      if (!adminAccess())
      {
         return -1;
      }

      SecurityCache sc = null;
      
      if (scope)
      {
         sc = getCache();
      }
      else
      {
         // Pick up the generation of security cache
         SecurityContextStack ctxs = SecurityContextStack.getContext();
         SecurityContext ctx = ctxs.getEffectiveContext();
         sc = ctx.getCache(); 
      }
      
      return sc.getSerial();
   }
   
   /**
    * Refreshes the security cache from the current state of the directory.
    * This is an admin API support method.
    *
    * @param    msgBuf
    *           list to receive messages if logmode enables admin messages
    * 
    * @return security cache generation serial number
    */
   public int adminRefresh(List<String[]> msgBuf)
   {
      // TO-DO granular check
      if (!adminAccess())
      {
         return -1;
      }

      // refreshing the security cache
      SecurityCacheRefresh ref = new SecurityCacheRefresh(
                                     SecurityCache.LM_DUAL, msgBuf);
      Thread thread;
      thread = new Thread(secThreads, ref, "refresh");

      thread.start();
      
      // wait till refresh is done
      try
      {
         thread.join();
      }
      catch (InterruptedException ie)
      {
         return -1;
      }

      return cacheSerial;
   }

   /**
    * Gets the parent scope for this scope.
    * Security Manager's scope has no parent.
    *
    * @return parent <code>Scope</code> object, which is <code>null</code>.
    */
   public final Scope getEnclosingScope()
   {
      return null;
   }

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   public void clientFinalize()
   {
   }

   /**
    * Gets the current debug level.
    * 
    * @return debug level. Levels < 0 disable debug output. Levels 0 and 
    *         higher enable debug output. The higher is the level, the more
    *         debug output is generated.
    */
   public Level getDebugLevel() 
   {
      return debugLevel; 
   }
   
   /**
    * Tests if logging is active for calls to {@link #dErr}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dErr} method.
    */
   public boolean isErrLevel()
   {
      return testDebugLevel(Level.SEVERE);
   }
      
   /**
    * Tests if logging is active for calls to {@link #dWarn}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dWarn} method.
    */
   public boolean isWarnLevel()
   {
      return testDebugLevel(Level.WARNING);
   }
      
   /**
    * Tests if logging is active for calls to {@link #dStat}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dStat} method.
    */
   public boolean isStatLevel()
   {
      return testDebugLevel(Level.INFO);
   }
      
   /**
    * Tests if logging is active for calls to {@link #dList}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dList} method.
    */
   public boolean isListLevel()
   {
      return testDebugLevel(Level.FINE);
   }
      
   /**
    * Tests if logging is active for calls to {@link #dTrace}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dTrace} method.
    */
   public boolean isTraceLevel()
   {
      return testDebugLevel(Level.FINER);
   }
      
   /**
    * Tests if logging is active for calls to {@link #dData}.
    *
    * @return   <code>true</code> if logging would occur in a call to the
    *           {@link #dData} method.
    */
   public boolean isDataLevel()
   {
      return testDebugLevel(Level.FINEST);
   }
      
   /**
    * Tests if the given logging level is active based on a comparison with
    * the currently configured logging level.
    *
    * @param    level
    *           The logging level to test.
    *
    * @return   <code>true</code> if logging would occur in a call with the
    *           given logging level.
    */
   public boolean testDebugLevel(Level level)
   {
      return level.intValue() >= debugLevel.intValue();
   }
      
   /**
    * Sets a new debug level.
    * <p>
    * Changing debug level is a protected action and requires the caller to
    * have sufficient access rights to the "debug" instance of the "system"
    * abstract resource.
    * 
    * @param  newLevel 
    *         debug level. Levels < 0 disable debug output. Levels 0 and 
    *         higher enable debug output. The higher is the level, the more
    *         debug output is generated.
    * @return <code>true</code> if the requested debug level was set
    */
   public synchronized boolean setDebugLevel(Level newLevel) 
   {
      boolean decision = true;
      
      // clients don't secure logging, but servers do
      if (cfg.isServer())
      {
         debugLevelRequest = newLevel;
         SystemResource sr = (SystemResource) getPluginInstance("system");
         decision = sr.checkDebugAccess();
         debugLevelRequest = NONE;
      }

      if (decision)
      {
         LOG.setLevel(newLevel);
         debugLevel = newLevel;
      }

      return decision; 
   }
   
   /**
    * Initializes and returns a valid SSL environment.
    *
    * @return  The SSL environment which has been configured with our
    *          custom key and trust stores.
    */
   public SSLContext getSecureSocketContext()
   throws NoSuchAlgorithmException,
          KeyStoreException, 
          UnrecoverableKeyException,
          KeyManagementException,
          ConfigurationException
   {
      return getSecureSocketContext(cfg);
   }
   
   /**
    * Initializes and returns a valid SSL environment.
    *
    * @param   config
    *          Bootstrap configuration information used to initialize the
    *          secure socket context.
    *          
    * @return   The SSL environment which has been configured with our
    *           custom key and trust stores.
    */
   public SSLContext getSecureSocketContext(BootstrapConfig config)
   throws NoSuchAlgorithmException,
          KeyStoreException, 
          UnrecoverableKeyException,
          KeyManagementException,
          ConfigurationException
   {
      SSLContext ctx = SSLContext.getInstance("TLS");

      // configure the SSL environment with our custom keys/certs
      TransportSecurity ts = getTransportSecurity(config);
      ts.attach(ctx);
      
      return ctx;
   }
   
   /**
    * Performs the requester side of a process or interactive client
    * authentication. Based on configuration, this method can validate the
    * server's certificate.
    * <p>
    * If required, the server's X.509 certificate is validated first. If this
    * does not fail, its DNS name is checked as defined in RFC2830, p3.6:
    * <ul>
    *   <li>if the certificate contains subjectAlternateNames of type dNSName,
    *       the set of those names is taken;
    *   <li>otherwise, the CN (common name) portion of the subject's DN is
    *       taken
    *   <li>the resulting set of names is compared agains the hostname used in
    *       establishing the TLS connection
    *   <li>'*' can be used as the leftmost component of the DNS name; it
    *       matches any name in that position
    * </ul>
    * If the DNS name check fails, an alternative check is done: the server's
    * certificate is compared with the certificate pointed to by the 
    * truststore alias. The comparison is done for encoded certificates and      
    * they have to be identical.
    * <p>
    * The next and last step is the client authentication procedure.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be 
    * <code>com.goldencode.p2j.net.SessionManager.createQueue()</code>.
    *
    * @param   socket
    *          The socket identifying this connection.
    * @param   config
    *          BootstrapConfig instance which should override the security
    *          manager's version;  <code>null</code> to use default config.
    *
    * @return  The data used to authenticate, or <code>null</code> if authentication failed.
    *
    * @throws  RestrictedUseException
    *          If called improperly.
    */
   public AuthData authenticateClient(NetSocket socket, BootstrapConfig config) 
   throws RestrictedUseException
   {
      // restrict the use of this method
      checkCallerAbort("com.goldencode.p2j.net.SessionManager.createQueue");
      
      if (config == null)
      {
         config = this.cfg;
      }
      
      if (config.isServer())
      {
         return null;
      }
      
      synchronized (SecurityManager.class)
      {
         return authenticateClientWorker(socket, config);
      }
   }
   
   /**
    * Does all the work as described above.
    * Temporarily uses the provided configuration instance.
    *
    * @param   socket
    *          The socket identifying this connection.
    * @param   config
    *          BootstrapConfig instance which should override the security
    *          Can't be <code>null</code>.
    *
    * @return  The data used to authenticate, or <code>null</code> if authentication failed.
    */
   private AuthData authenticateClientWorker(NetSocket socket, BootstrapConfig config) 
   {
      boolean isProcess = false;
      String remoteAuthID = null;
      String appServer = "";

      ObjectInputStream  in  = socket.getObjectInput();
      ObjectOutputStream out = socket.getObjectOutput();

      boolean secure = config.getBoolean("net", "connection", "secure", false);
      
      if (isTraceLevel())
      {
         dTrace(String.format("secure = %b", secure));
      }
      
      // certificate processing
      
      SSLSession locSess = socket.getSession();
      
      if (locSess == null && secure)
      {
         dTrace("Secure mode requires a certificate sent, which is missing.");
         socket.close();
         return null;
      }

      // query local certificate
      java.security.cert.Certificate[] locCerts = null;
      
      if (secure)
      {
         locCerts = locSess.getLocalCertificates();
      }
      
      boolean sentCert = (locCerts != null && locCerts.length > 0);
      
      if (isTraceLevel())
      {
         dTrace(String.format("sentCert = %b", secure));
      }
      
      // validate the server's certificate if requested
      if (secure && !validateServer(locSess, config))
      {
         dTrace("Server certificate validation failed.");
         socket.close();
         return null;
      }

      if (!sentCert)
      {
         // state the client's intention in case of anonymous connection
         if (!sendAuthType(out, config))
         {
            dTrace("Anonymous connection refused.");
            socket.close();
            return null;
         }
      }
      
      // retryable section of the authentication protocol

      int authModeReq = AUTH_MODE_NONE;
      int authResult  = AUTH_RESULT_NONE;
      int authAction  = AUTH_ACTION_CONTINUE;
      
      Authenticator auth     = null;
      String        hookParm = null;

      try
      {
         retryLoop:
         while (authAction == AUTH_ACTION_CONTINUE ||
                authAction == AUTH_ACTION_RETRY)
         {
            dTrace("Getting AUTHMODEREQ");
            
            // read the AUTHMODEREQ packet from the server
            try
            {
               authModeReq = in.readInt();
               authAction  = in.readInt();
            }
            
            catch (IOException ioe)
            {
               dTrace("Error getting AUTHMODEREQ " + ioe);
               authAction = AUTH_ACTION_ABORT;
            }
            
            dTrace("Got AUTHMODEREQ: " + authModeReq + " (action = " +
                   authAction + ")");
            
            // interpret AUTHMODEREQ packet contents
            switch (authAction)
            {
               case AUTH_ACTION_ABORT:
                  dTrace("Aborting as requested.");
                  socket.close();
                  return null;
                  
               case AUTH_ACTION_DONE:
                  dTrace("Done with retry loop.");
                  break retryLoop;
            }
            
            byte[] authData = null;
            
            // some auth mode types are not done yet, execute next phase
            switch (authModeReq)
            {
               case AUTH_MODE_X509:
                  // This mode means the client's certificate was required
                  // to be sent previously (at TLS connection time) but it 
                  // was not sent.  This is a protocol logic issue since we
                  // should have aborted above.
                  dTrace("Missing certificate (protocol error).");
                  socket.close();
                  return null;
      
               case AUTH_MODE_IDPW:
                  // This mode means that the server wants ID and password and
                  // does not care about client's certificate whether it was
                  // sent or not, fall-through.
      
               case AUTH_MODE_X509_IDPW:
                  // This mode means that the server wants ID and password as
                  // well as the client's certificate which was sent (otherwise
                  // AUTH_RESULT_FAILURE would be received).
                  authData = clientAuthHookWorker(null, authResult, config);
                  if (authData == null)
                  {
                     dTrace("clientAuthHookWorker failed.");
                     socket.close();
                     return null;
                  }

                  auth = this;
                  break;
                  
               case AUTH_MODE_CUSTOM:
                  // This mode means that the server runs custom
                  // authentication hook regardless of the certificate. The
                  // call below may return null.  This typically loads a
                  // class via data sent on the connection and returns an
                  // instance of that class.  Note this only happens the
                  // first time we come through).
                  if (authResult == AUTH_RESULT_NONE)
                  {
                     // reads the first 2 pieces of the packet, loads the
                     // class and returns an instance
                     auth = obtainCustomClientHook(in);
                     try
                     {
                        // the packet is made of three pieces:
                        // read piece 3 - class parameters
                        hookParm = in.readUTF();
                     }
                     catch (IOException io)
                     {
                        dTrace("Problem reading hook parm " + io);
                        hookParm = "";
                     }
                  }
   
                  if (auth != null)
                  {
                     Map<String, Object> parameters =
                                            new HashMap<String, Object>();
                     parameters.put("option", hookParm);
                     try
                     {
                        parameters.put("subjectId",
                               cfg.getConfigItem("access", "subject", "id"));
                     }
                     catch (ConfigurationException e)
                     {
                        // ignore
                     }

                     authData = auth.clientAuthHook(parameters, authResult);
                  }
                  else
                  {
                     dTrace("Custom client auth hook failed.");
                     socket.close();
                     return null;
                  }
                  break;
            }
      
            // provide authentication data to the server
            int authdl = authData.length;
            
            dTrace("Sending AUTHDATA");
            try
            {
               // send the packet
               out.writeInt(authdl);
               if (authdl > 0)
               {
                  out.write(authData);
               }
               out.flush();
            }
            catch (IOException ioe)
            {
               dTrace("Failure sending AUTHDATA " + ioe);
               socket.close();
               return null;
            }
            dTrace("Sent AUTHDATA");
            
            // after notifying the server with the authdata, check if
            // the plugin wants to exit
            if (authdl == 0)
            {
               dTrace("Plugin is forcing system exit now.");
               System.exit(0);
            }
      
            authResult = AUTH_RESULT_NONE;
      
            dTrace("Getting AUTHRESULT");
            
            // receive AUTHRESULT packet
            try
            {
               authResult = in.readInt();
               authAction = in.readInt();
            }
            
            catch (IOException ioe)
            {
               dTrace("Failure reading AUTHRESULT " + ioe);
               authAction = AUTH_ACTION_ABORT;
            }
            
            dTrace("Got AUTHRESULT: " + authResult + " (action = " +
                   authAction + ")");
            
            if (authAction == AUTH_ACTION_ABORT)
            {
               dTrace("AUTHRESULT forced abort.");
               socket.close();
               return null;
            }
         }

         // if all is fine, then read the subject ID determined on the
         // remote side
         
         // read the authentication ID, so that it can be saved
         try
         {
            appServer = (String) in.readObject();
            isProcess = in.readBoolean();
            remoteAuthID = in.readUTF();
         }
         
         catch (ClassNotFoundException cex)
         {
            dTrace("Failure reading appServer name " + cex);
            authAction = AUTH_ACTION_ABORT;
         }
         
         catch (IOException ioe)
         {
            dTrace("Failure reading AUTHRESULT " + ioe);
            authAction = AUTH_ACTION_ABORT;
         }
         
         if (authAction == AUTH_ACTION_ABORT)
         {
            dTrace("AUTHRESULT forced abort.");
            socket.close();
            return null;
         }
      }
         
      finally
      {
         if (auth != null)
            auth.clientFinalize();
      }
      
      // return the identity used to authenticate
      return new AuthData(isProcess, remoteAuthID, appServer); 
   }

   /**
    * Implements the server side of the authentication procedure for
    * processes or interactive clients.  Optionally verifies client's
    * credentials.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be 
    * <code>com.goldencode.p2j.net.RouterSessionManager$Incoming.run()</code>.
    * <p>
    * There are two major classes of connections: certified and anonymous.
    *
    * @param    socket
    *           Socket identifying this connection.
    *
    * @return   Reference to an object serving as the session key if 
    *           authenticated, otherwise either <code>null</code> for failures
    *           or a string describing normal client exit from authentication.
    *
    * @throws   RestrictedUseException
    *           If called improperly.
    */
   public Object authenticateLocal(NetSocket socket) 
   throws RestrictedUseException
   {
      // restrict the use of this method
      checkCallerAbort(
         "com.goldencode.p2j.net.RouterSessionManager$Incoming.run");
      
      if (!cfg.isServer())
         return null;

      ObjectInputStream  in  = socket.getObjectInput();
      ObjectOutputStream out = socket.getObjectOutput();

      SecurityCache sc = getCache(); 
      
      // safely query cached data
      int    authMode    = sc.getAuthMode();
      int    authRetries = sc.getAuthNumRetries();
      String anonymId    = sc.getAnonymId();
      String hookName    = sc.getHookName();

      // process TLS session certificates

      SSLSession locSess = socket.getSession();
      boolean recvCert = false;
      String alias = null;
      Account[] accs = null;

      java.security.cert.Certificate[] peerCerts = null;
      
      try
      {
         // if in secure mode
         if (locSess != null)
         {
            // obtain the certs from the other side
            peerCerts = locSess.getPeerCertificates();
         }
      }
      catch (Exception e)
      {
         // exception is handled as if there was 0 certificates received
      }
      
      if (peerCerts != null && peerCerts.length > 0)
      {
         X509Certificate cert = (X509Certificate)peerCerts[0];
         try
         {
            recvCert = true;  
            dTrace("Received certificate of " + getSubjectCommonName(cert));
         }
         catch (CertificateParsingException e)
         {
            // this exception can't be tolerated
            return null;
         }
      }
      
      // deal with TLS connection type

      String localUserId = null;
      String localProcId = null;
      
      int authModeReq = authMode;
      int authAction  = AUTH_ACTION_CONTINUE;
      int authType    = -1;
      
      // the following code will reset the authModeReq to be a return code
      // in any case where the result is actually known without needing to
      // contact the client again, if this is not reset, then it represents
      // the authentication mode which the client must enforce
      
      if (!recvCert)
      {
         // anonymous connection: the client has to tell the type of
         // the connection. allowed types are: 
         // - process
         //   will be accepted if there is an anonympus process account
         // - interactive
         //   the interactive client will be instructed by the server
         //   about the server requested authentication mode
         // - programmatic
         //   not a regular interactive client (web applet for instance);
         //   will be accepted as AUTH_MODE_IDPW but the user account
         //   has to explicitly allow this authentication type AND
         //   some extra protection should be implemented between the web
         //   server and the security manager (one time token alike)

         dTrace((locSess == null) ? "Insecure socket connection."
                                  : "TLS connection without certificate.");
   
         // get the suggested authentication type from the client
                                  
         String requestedUserName = null;
         dTrace("Getting AUTHTYPE");
         try
         {
            // read the packet
            authType = in.readInt();

            if (authType == AUTH_REQ_USER)
            {
               requestedUserName = in.readUTF();
               if (requestedUserName.equals(""))
                  requestedUserName = null;

               if (requestedUserName != null)
               {
                  Account acc = sc.getAccountById(requestedUserName);
                  if (acc != null && acc instanceof UserAccount)
                  {
                     // effective mode will be used
                     UserAccount userAccount = ((UserAccount) acc);
                     authMode = userAccount.getAuthMode();
                     hookName = userAccount.getAuthPlugin(); 
                     authModeReq = authMode;
                  }
               }
            }
         }

         catch (SocketException se)
         {
            dTrace("Error getting AUTHTYPE " + se);
            return null;
         }

         catch (IOException ioe)
         {
            dErr("Error getting AUTHTYPE " + ioe);
            return null;
         }
         dTrace("Got AUTHTYPE: " + authType + " userName: " + requestedUserName);
         
         // verify the suggested authentication type
         if (authType == AUTH_REQ_PROCESS)
         {
            // process types are only allowed if there is an associated
            // anonymous process id that is defined
            if (anonymId == null)
            {
               authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
               authAction  = AUTH_ACTION_ABORT;
            }
            else
            {
               authModeReq = AUTH_RESULT_SUCCESS;
               authAction  = AUTH_ACTION_DONE;
               localProcId = anonymId;
            }
         }
         else if (authType == AUTH_REQ_USER)
         {
            // this is the regular interactive user over an anonymous link
            // fall-through
         }
         else if (authType == AUTH_REQ_PROGRAM)
         {
            // programmatic interactive (applet) access
            authMode    = AUTH_MODE_IDPW;
            authModeReq = AUTH_MODE_IDPW;
         }
         else
         {
            // unknown authentication type
            authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
            authAction  = AUTH_ACTION_ABORT;
         }
      }
      else
      {
         // certified connection, identify the account by certificate and
         // figure out truststore alias of the certificate
         try
         {
            alias = sc.getTrustStore().getCertificateAlias(peerCerts[0]);
         }
         catch (KeyStoreException e)
         {
            // nothing to do
         }

         if (alias == null)
         {
            dTrace("Can't identify the alias");
            authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
            authAction  = AUTH_ACTION_ABORT;
         }
         else
         {
            dTrace("Certificate has alias <" + alias + ">");

            int procs = sc.getCountByAlias(alias, Account.ACC_PROCESS);
            int users = sc.getCountByAlias(alias, Account.ACC_USER);
            dTrace("Associated process accounts: " + procs);
            dTrace("Associated user    accounts: " + users);

            if (procs + users == 0)
            {
               dTrace("No account for alias <" + alias + ">");
               authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
               authAction  = AUTH_ACTION_ABORT;
            }
            else
            {
               if (procs > 0 && users > 0)
               {
                  dTrace("Mixed user and process accounts for alias <" + 
                         alias + ">");
                  authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
                  authAction  = AUTH_ACTION_ABORT;
               }
               else
               {
                  if (procs > 0)
                  {
                     // processes only
                     if (procs > 1)
                     {
                        dTrace("Ambiguous association of process accounts"); 
                        authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
                        authAction  = AUTH_ACTION_ABORT;
                     }
                     else
                     {
                        accs = sc.getAccountsByAlias(alias, 
                                                     Account.ACC_PROCESS);
                        localProcId = accs[0].getSubjectId();
                        dTrace("Associated process account is <" + 
                               localProcId + ">");
                        authModeReq = AUTH_RESULT_SUCCESS;
                        authAction  = AUTH_ACTION_DONE;
                     }
                  }
                  else
                  {
                     // users only, this is a table of valid user IDs
                     accs = sc.getAccountsByAlias(alias, Account.ACC_USER);
                     if (accs != null && accs.length == 1)
                     {
                        // effective mode will be used
                        authModeReq = ((UserAccount) accs[0]).getAuthMode();
                     }
                  }
               }
            }
         }
      }

      // retryable section of the authentication protocol

      int authResult = AUTH_RESULT_NONE;

      do
      {
         //
         // interpreting AUTHMODE
         //
         
         switch (authModeReq)
         {
            case AUTH_MODE_X509:
               // this mode means the client's certificate is required
               if (!recvCert)
               {
                  // certificate required but not received
                  authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
                  authAction  = AUTH_ACTION_ABORT;
                  break;
               }
               // The user certificate has been identified. One or more
               // accounts are selected. This mode required an unambiguous
               // association of certificate (alias) to an user account. If
               // only one account has been selected, it gets used. Otherwise,
               // the set has to include a special default account where 
               // subjectID == alias.
               if (accs.length > 1)
               {
                  int i;
                  for (i = 0; i < accs.length; i++)
                  {
                     String subid = accs[i].getSubjectId();
                     if (subid.compareToIgnoreCase(alias) == 0)
                     {
                        localUserId = accs[i].getSubjectId();
                        authModeReq = AUTH_RESULT_SUCCESS;
                        authAction  = AUTH_ACTION_DONE;
                        break;
                     }
                  }
                  if (i == accs.length)
                  {
                     dTrace("Ambiguous association of user accounts"); 
                     authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
                     authAction  = AUTH_ACTION_ABORT;
                  }
               }
               else
               {
                  localUserId = accs[0].getSubjectId();
                  authModeReq = AUTH_RESULT_SUCCESS;
                  authAction  = AUTH_ACTION_DONE;
               }
               break;
   
            case AUTH_MODE_IDPW:
               // This mode means that the server wants ID and password and
               // does not care about client's certificate whether it was sent
               // or not. Drop all preselected accounts.
               accs = null; 
               authModeReq = authMode;
               break;
   
            case AUTH_MODE_X509_IDPW:
               // This mode means that the server wants ID and password as
               // well as the client's certificate which should have been
               // sent.
               if (!recvCert)
               {
                  // certificate required but not received
                  authModeReq = AUTH_RESULT_UNSPECIFIED_FAILURE;
                  authAction  = AUTH_ACTION_ABORT;
                  break;
               }
               // The user certificate has been identified. One or more
               // accounts were preselected. The user sends a preferred ID
               // with password.
               authModeReq = authMode;
               break;
               
            case AUTH_MODE_CUSTOM:
               // This mode means that the server runs custom authentication
               // hook regardless.
               authModeReq = authMode;
               break;
         }
   
         if (authModeReq != AUTH_RESULT_SUCCESS)
         {  
            dTrace("Sending AUTHMODEREQ: " + authModeReq + " (action = " +
                   authAction + ")");
            
            // send the AUTHMODEREQ packet to the client
            try
            {
               // send the packet
               out.writeInt(authModeReq);
               out.writeInt(authAction);
               out.flush();
            }
            
            catch (SocketException se)
            {
               dTrace("Error sending AUTHMODEREQ " + se);
               return null;
            }

            catch (IOException ioe)
            {
               dErr("Error sending AUTHMODEREQ " + ioe);
               return null;
            }
            
            dTrace("Sent AUTHMODEREQ");

            String serverHookName = null;

            // only send the class down the first time
            if (authModeReq == AUTH_MODE_CUSTOM &&
                authResult == AUTH_RESULT_NONE)
            {
               // get a copy of the hook class bytes, try name and nameClient
               String clientHookName = hookName;
               serverHookName = hookName;
               byte[] classData = getClassBytes(hookName);
               if (classData == null)
               {
                  // hook probably consists of separate client- and
                  // server-side classes
                  dList("no hook class for " + clientHookName);
                  clientHookName = hookName + "Client";
                  classData = getClassBytes(clientHookName);
                  if (classData == null)
                  {
                     dWarn("no hook class for " + clientHookName);
                     clientHookName = hookName;
                  }
                  else
                  {
                     dList("hook class is " + clientHookName);
                     serverHookName = hookName + "Server";
                     
                     if (!classExists(serverHookName))
                     {
                        // use the server's default processing as there is
                        // no user-replaceable server-side plugin
                        serverHookName = null;
                     }
                  }
               }
               else
                  dList("hook class is " + clientHookName);

               int csize = classData == null ? 0 : classData.length;
            
               // send AUTHCUSTOM packet
               dTrace("Sending AUTHCUSTOM");
               try
               {
                  // the packet is made of three pieces:
                  // 1 - byte array, copy of the class to run
                  //     formatted as int + bytes, where int == 0 means no class
                  // 2 - fully qualified class name formatted as UTF
                  // 3 - string of class parameters formatted as UTF
   
                  // write piece 1 - may be empty
                  out.writeInt(csize);
                  if (csize > 0)
                     out.write(classData);
   
                  // write piece 2 - class name
                  out.writeUTF(clientHookName);
   
                  // write piece 3 - class parameters
                  String hookParm = getHookParam(sc, hookName);
                  if (hookParm == null)
                     hookParm = "";
                  out.writeUTF(hookParm);
   
                  out.flush();
               }

               catch (SocketException se)
               {
                  dTrace("Error sending AUTHCUSTOM " + se);
                  return null;
               }

               catch (IOException ioe)
               {
                  dErr("Error sending AUTHCUSTOM " + ioe);
                  return null;
               }
               dTrace("Sent AUTHCUSTOM");
            }
      
            if (authAction == AUTH_ACTION_ABORT)
            {
               return null;
            }
            
            byte authData[] = null;
            int  authSize = 0;
      
            dTrace("Getting AUTHDATA");
            
            // get authentication data from the client
            try
            {
               authSize = in.readInt();
               dTrace("Getting AUTHDATA authSize="+authSize);
               
               if (authSize == 0)
               {
                  // special case - client want to exit authentication cleanly
                  return "client decided to quit authentication";
               }
               
               authData = new byte[authSize];
               in.readFully(authData);
            }
            catch (IOException ioe)
            {
               dTrace("Error getting AUTHDATA " + ioe);
               return null;
            }
            dTrace("Got AUTHDATA");
      
            AuthenticationResponse response = null;
            authResult = AUTH_RESULT_NONE;
            
            // process authentication data
            switch (authMode)
            {
               case AUTH_MODE_IDPW:
               case AUTH_MODE_X509_IDPW:
                  response    = serverAuthHook(authData, "");
                  localUserId = response.getSubject();
                  authResult  = response.getReturnCode();
 
                  if (authResult != AUTH_RESULT_SUCCESS)
                  {
                     break;
                  }
                  
                  // for programmatic call, check the account for matching
                  // authentication type
                  if (authType == AUTH_REQ_PROGRAM)
                  {   
                     UserAccount user = (UserAccount)sc.
                                        getAccountById(localUserId);
                     if (user.getAuthMode() != AUTH_MODE_IDPW)
                     {
                        dTrace("No programmatic use for ID <" + localUserId + 
                               ">");
                        localUserId = null;
                     }
                  }
                  // check the requested ID against the list of allowed IDs
                  else if (accs != null)
                  {
                     int i;
                     
                     for (i = 0; i < accs.length; i++)
                     {
                        if (accs[i].getSubjectId()
                                   .compareToIgnoreCase(localUserId) == 0)
                           break;
                     }
                     
                     if (i == accs.length)
                     {
                        dTrace("Not allowed to use account <" + localUserId + 
                               ">"); 
                        localUserId = null;
                     }
                  }
                  break;
                  
               case AUTH_MODE_CUSTOM:
                  // this mode means that the server runs custom auth hook
                  response = runCustomServerHook(sc,
                                                 hookName,
                                                 serverHookName,
                                                 getHookParam(sc, hookName),
                                                 authData); 
                  localUserId = response.getSubject();
                  authResult  = response.getReturnCode(); 
                  break;
            }
      
            // send the negative AUTHRESULT packet to the client
            if (localUserId == null)
            {
               authAction = AUTH_ACTION_ABORT;
               
               if (authRetries == -1 || authRetries > 0)
               {
                  authAction = AUTH_ACTION_RETRY;
                  
                  if (authRetries > 0)
                  {
                     authRetries--;
                  }
               }
      
               if (!sendAuthResult(out, authResult, authAction))
               {
                  return null;
               }

               if (authAction == AUTH_ACTION_ABORT)
               {
                  return null;
               }
            }
            else
            {
               authAction = AUTH_ACTION_DONE;
            }
         }
         
         // reset to the default expected mode
         authModeReq = authMode;
         
      } while (authAction == AUTH_ACTION_RETRY);
                                            
      // check for the enabled status of the accounts
      boolean allow = true;
      String appServer = null;
      
      if (localProcId != null)
      {
         ProcessAccount pa = (ProcessAccount) sc.getAccountById(localProcId);
         if (pa == null)
         {
            allow = false;
         }
         else
         {
           allow = pa.isEnabled();
           appServer = pa.getAppServer();
         }
      }
      
      if (allow && localUserId != null)
      {
         UserAccount ua = (UserAccount) sc.getAccountById(localUserId);
         if (ua == null)
         {
            allow = false;
         }
         else
         {
           allow = ua.isEnabled();
           appServer = ua.getAppServer();
         }
      }
      
      if (!allow)
      {
         dTrace("Account disabled");
         authResult = AUTH_RESULT_INSUFFICIENT_RIGHTS;
         authAction = AUTH_ACTION_ABORT;
         sendAuthResult(out, authResult, authAction);
         return null;
      }
      
      authResult = AUTH_RESULT_SUCCESS;
      authAction = AUTH_ACTION_DONE;

      // create new security context
      SecurityContext newContext = createSecurityContext(sc,
                                                         localProcId,
                                                         localUserId);
      
      if (newContext == null || sc != getCache())
      {
         dTrace("authenticateLocal: newer security cache generation available; " +
                "authentication aborted.");

         authResult = AUTH_RESULT_UNSPECIFIED_FAILURE;
         authAction = AUTH_ACTION_ABORT;
         sendAuthResult(out, authResult, authAction);
         return null;
      }
      
      // create new session
      Session sess = new Session(socket, sc, newContext);
      
      synchronized(listOfSess)
      {
         listOfSess.add(sess);
      }
      
      assignContext(newContext);

      // check access rights to the system/logon resource
      pushContextWorker(newContext);
      SystemResource sr = (SystemResource) getPluginInstance("system");
      boolean decision = sr.checkLogonAccess();
      popContextWorker();
      
      if (!decision)
      {
         dTrace("Access denied to system/logon resource");
         authResult = AUTH_RESULT_INSUFFICIENT_RIGHTS;
         authAction = AUTH_ACTION_ABORT;
      }

      // send the AUTHRESULT packet to the client
      if (!sendAuthResult(out, authResult, authAction))
      {
         // error sending the result; do not keep the session
         decision = false;
      }
      else
      {
         // send the account used for authentication back to the requester
         // so that it can save it.
         dTrace("Sending LOCALUSERID: " + localUserId);
        
         // send the LOCALUSERID packet to the client
         try
         {
            // send the packet
            out.writeObject(appServer);
            out.writeBoolean(localUserId == null && localProcId != null);
            out.writeUTF(localUserId != null ? localUserId : localProcId);
            out.flush();
         }
         
         catch (SocketException se)
         {
            dTrace("Error sending LOCALUSERID " + se);
            decision = false;
         }
   
         catch (IOException ioe)
         {
            dErr("Error sending LOCALUSERID " + ioe);
            decision = false;
         }
      }

      if (!decision)
      {
         synchronized (listOfSess)
         {
            listOfSess.remove(sess);
         }
         return null;
      }

      return newContext;
   }

   /**
    * Returns the subject ID associated with the current context.
    * <p>
    * The subject ID returned is either userID, or processId, if no user is
    * logged on.
    *
    * @return subject ID or "" if nothing is established
    */
   public String getUserId() 
   {
      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // figure out subject ordinal
      int userOrd = ctx.getUserOrdinal();
      if (userOrd == -1)
         userOrd = ctx.getProcessOrdinal();
      if (userOrd == -1)
         return "";

      // return the subject's identity
      return sc.getAccountByOrd(userOrd).getSubjectId();
   }

   /**
    * Returns the the appserver associated with the current context.
    * <p>
    * If no appserver is associated with the current context, it returns <code>null</code>
    *
    * @return   See above.
    */
   public String getAppServer() 
   {
      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // figure out subject ordinal
      int userOrd = ctx.getUserOrdinal();
      if (userOrd == -1)
         userOrd = ctx.getProcessOrdinal();
      if (userOrd == -1)
         return null;

      // return the associated appserver
      return sc.getAccountByOrd(userOrd).getAppServer();
   }

   /**
    * Reset the effective security context associated with the current thread.
    * 
    * @throws   RestrictedUseException
    *           If called outside of expected code. 
    */
   public void resetContext()
   throws RestrictedUseException
   {
      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.util.Agent$ResetContextCommand.execute"
      };
      checkCallerAbort(callers);

      SecurityContextStack.getContext().getEffectiveContext().reset();
   }

   /**
    * Detects if there is a valid context associated with the current thread.
    *
    * @return   <code>true</code> if there is a valid security context.
    */
   public boolean hasContext()
   {
      return SecurityContextStack.getContext() != null;
   }
   
   /**
    * Get the effective security context associated with the current thread,
    * if any.
    * 
    * @return  Effective security context, or <code>null</code> if either
    *          this is not a server-side security manager, or there is no
    *          effective security context associated with the current
    *          thread.
    *          
    * @throws  RestrictedUseException
    *          if called improperly.
    */
   public Object getEffectiveContext()
   throws RestrictedUseException
   {
      if (!cfg.isServer())
      {
         return null;
      }
      
      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.net.RouterSessionManager.connectVirtual", 
      };
      checkCallerAbort(callers);
      
      SecurityContextStack scs = SecurityContextStack.getContext();
      
      return (scs != null ? scs.getEffectiveContext() : null);
   }
   
   /**
    * Get the number which uniquely identifies the current context, or
    * <code>null</code> if there is no current context.
    *
    * @return  Current security context's session ID.
    */
   public Integer getSessionId()
   {
      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      
      return (ctx != null ? (new Integer(ctx.getSessionId())) : null);
   }
   
   /**
    * Get the session token which uniquely identifies the current context, or
    * <code>null</code> if there is no current context.
    *
    * @return  Current security context's session token.
    */
   public SessionToken getSessionToken()
   {
      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      
      return (ctx != null ? ctx.getSessionToken() : null);
   }
   
   /**
    * Get the identifier mapped to the current thread.  This is only
    * meaningful for a given unit of work, since threads generally are
    * recycled into a thread pool (and their thread ID unmapped) when their
    * unit of work is done.
    *
    * @return  Current thread's ID.
    */
   public Integer getThreadId()
   {
      synchronized (threadIDs)
      {
         return (Integer) threadIDs.get(Thread.currentThread());
      }
   }

   /**
    * Returns ALL subject IDs associated with the current context.
    * <p>
    * The first subject ID in the array is either a userID or a processId
    * (if no user is logged on).  All subsequent subject IDs are the list
    * of group names.  Note that as long as there is a security context,
    * there will be at least 1 subject ID.  The group subject IDs will only
    * be there if the user or process belongs to 1 or more groups.
    *
    * @return   The array of subject IDs or an empty array (length == 0) if
    *           no security context exists.
    */
   public String[] getAccountIds() 
   {
      // find the security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext      ctx  = ctxs.getEffectiveContext();
      SecurityCache        sc   = ctx.getCache(); 

      // result list (initially empty)
      String[] list = new String[0];      
      
      int[] groupList = null;
      
      // figure out 1st subject ordinal
      int ord = ctx.getUserOrdinal();
      
      if (ord == -1)
      {
         // we aren't a user, perhaps we are a process?
         ord = ctx.getProcessOrdinal();
      }
      else
      {
         UserAccount ua = (UserAccount) sc.getAccountByOrd(ord); 
         groupList = ua.getGroups();
      }
      
      if (ord != -1)
      {
         // we have at least 1 subject ID to return, store it
         ArrayList array = new ArrayList();
         array.add(sc.getAccountByOrd(ord).getSubjectId());
         
         // add groups
         if (groupList != null)
         {
            for (int i = 0; i < groupList.length; i++)
            {
               array.add(sc.getAccountByOrd(groupList[i]).getSubjectId());
            }
         }
         
         // convert to an array
         list = (String[]) array.toArray(list);
      }

      // return the subject list
      return list;
   }

   /**
    * Returns the subject ID associated with the passed key.
    * <p>
    * The subject ID returned is in the form processID/userID. Either part can
    * be empty.
    * No group IDs are returned as they must be uniquely identifiable by
    * userID.
    * <p>
    * Note that the security cache has to be taken from the passed key which
    * is a SecurityContext, as the correct identity of the key is bound to the
    * security cache generation.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be 
    * the "authenticateRemote" method of the
    * com.goldencode.p2j.net.Dispatcher class.
    *
    * @param  key
    *         a key to a security context to be queried
    *
    * @return subject ID 
    *
    * @throws RestrictedUseException if called improperly
    */
   public String getIdentity(Object key) 
   throws RestrictedUseException
   {
      // restrict the use of this method
      checkCallerAbort(
         "com.goldencode.p2j.net.RouterSessionManager.connectVirtual");
      
      if (!cfg.isServer())
      {
         return null;
      }
      
      // check the passed object
      if (!(key instanceof SecurityContext))
      {
         throw new RestrictedUseException("invalid security context key");
      }

      // Pick up the generation of security cache
      SecurityContext ctx = (SecurityContext)key;
      SecurityCache sc = ctx.getCache(); 

      int procOrd = ctx.getProcessOrdinal();
      String procId = null;
      if (procOrd != -1)
         procId = sc.getAccountByOrd(procOrd).getSubjectId();
      else
         procId = "";

      int userOrd = ctx.getUserOrdinal();
      String userId = null;
      if (userOrd != -1)
         userId = sc.getAccountByOrd(userOrd).getSubjectId();
      else
         userId = "";

      return new String(procId + "/" + userId);
   }

   /**
    * Implements the authentication procedure for interactive clients which
    * run inside the same JVM process as the rest of the server.
    * <p>
    * <b>Restricted Use</b>. This method checks the caller to be 
    * <code>com.goldencode.p2j.main.ClientCore.start()</code>.
    *
    * @param    cfg
    *           Local configuration.
    *
    * @return   Reference to an object serving as the session key if 
    *           authenticated, otherwise either <code>null</code> for failures
    *           or a string describing normal client exit from authentication.
    *
    * @throws   RestrictedUseException
    *           If called improperly.
    */
   public Object authenticateSingle(BootstrapConfig cfg) 
   throws RestrictedUseException
   {
      // restrict the use of this method
      checkCallerAbort("com.goldencode.p2j.main.ClientCore.start");
      
      if (!cfg.isServer())
         return null;

      SecurityCache sc = getCache(); 
      
      // safely query cached data
      int    authMode    = sc.getAuthMode();
      int    authRetries = sc.getAuthNumRetries();
      String anonymId    = sc.getAnonymId();
      String hookName    = sc.getHookName();
      int    authModeReq = authMode;
      int    authAction  = AUTH_ACTION_CONTINUE;
      int    authResult  = AUTH_RESULT_NONE;
      int    authType    = -1;
      String userid      = null;
      
      // get the suggested authentication type from the client
      String requestedId = cfg.getString("access", "subject", "id", null);
                                          
      if (requestedId != null)
      {
         Account acc = sc.getAccountById(requestedId);
         
         // override from account specific settings
         if (acc != null && acc instanceof UserAccount)
         {
            // effective mode will be used
            UserAccount userAccount = ((UserAccount) acc);
            authMode = userAccount.getAuthMode();
            hookName = userAccount.getAuthPlugin(); 
         }
      }
      
      // retryable section of the authentication protocol
      
      do
      {
         authModeReq = authMode;
         
         // try the common hook class using the base name
         String srvHook = hookName;
         Authenticator auth = obtainCustomAuthHook(hookName);
         
         if (auth == null)
         {
            srvHook = null;
            
            // fallback to the split hook class
            auth = obtainCustomAuthHook(hookName + "Client");
            
            if (auth != null)
            {
               if (classExists(hookName + "Server"))
               {
                  srvHook = hookName + "Server";
               }
               
               hookName = hookName + "Client";
            }
            else
            {
               // fallback to default
               auth = new DefaultLoginPanel();
            }
         }
         
         String hookParm = getHookParam(sc, hookName);
         
         if (hookParm == null)
            hookParm = "";
         
         Map<String, Object> parameters = new HashMap<String, Object>();
         parameters.put("option", hookParm);
         parameters.put("subjectId",
                        cfg.getString("access", "subject", "id", null));

         byte[] authData = auth.clientAuthHook(parameters, authResult);
         
         
         AuthenticationResponse response = null;
         authResult = AUTH_RESULT_NONE;
         
         // process authentication data
         switch (authMode)
         {
            case AUTH_MODE_IDPW:
               response    = serverAuthHook(authData, "");
               userid      = response.getSubject();
               authResult  = response.getReturnCode();
               break;
               
            case AUTH_MODE_CUSTOM:
               // this mode means that the server runs custom auth hook
               response = runCustomServerHook(sc,
                                              hookName,
                                              srvHook,
                                              getHookParam(sc, hookName),
                                              authData); 
               userid      = response.getSubject();
               authResult  = response.getReturnCode(); 
               break;
         }
   
         // send the negative AUTHRESULT packet to the client
         if (userid == null)
         {
            authAction = AUTH_ACTION_ABORT;
            
            if (authRetries == -1 || authRetries > 0)
            {
               authAction = AUTH_ACTION_RETRY;
               
               if (authRetries > 0)
               {
                  authRetries--;
               }
            }
   
            if (authAction == AUTH_ACTION_ABORT)
            {
               return null;
            }
         }
         else
         {
            authAction = AUTH_ACTION_DONE;
         }
         
         // reset to the default expected mode
         authModeReq = authMode;
         
      } while (authAction == AUTH_ACTION_RETRY);
                                            
      // check for the enabled status of the account
      if (userid != null)
      {
         UserAccount ua = (UserAccount) sc.getAccountById(userid);
         
         if (ua == null || !ua.isEnabled())
         {
            dTrace("Account disabled");
            return null;
         }
      }
      
      authResult = AUTH_RESULT_SUCCESS;
      authAction = AUTH_ACTION_DONE;

      SecurityContextStack cts = SecurityContextStack.getContext();
      
      // create new security context and move all temporary tokens into it
      SecurityContext oldContext = cts.getInitial();
      SecurityContext newContext = createSecurityContext(sc, null, userid);
      
      if (newContext == null || sc != getCache())
      {
         dTrace("authenticateSingle: newer security cache generation available; " +
                "authentication aborted.");

         authResult = AUTH_RESULT_UNSPECIFIED_FAILURE;
         authAction = AUTH_ACTION_ABORT;
         return null;
      }

      oldContext.transferAllTokens(newContext);
      
      // create new session
      Session sess = new Session(Thread.currentThread(), sc, newContext);
      
      synchronized(listOfSess)
      {
         listOfSess.add(sess);
      }
      
      // switch to new context on this thread
      // assignContext(newContext);
      pushContextWorker(newContext);

      SystemResource sr = (SystemResource) getPluginInstance("system");
      
      // check access rights to the system/logon resource
      if (!sr.checkLogonAccess())
      {
         // clear context
         popContextWorker();
         
         dTrace("Access denied to system/logon resource");
         
         synchronized (listOfSess)
         {
            listOfSess.remove(sess);
         }
         
         return null;
      }

      return newContext;
   }

   /**
    * Verifies peer's credentials and performs remote authentication.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be 
    * <code>com.goldencode.p2j.net.RouterSessionManager.authenticateRemote()</code>.
    *
    * @param  identity
    *         remote application/client identity in the form "processID/userID"
    *
    * @return reference to an object serving as a key if authenticated, 
    *         otherwise <code>null</code>
    *
    * @throws RestrictedUseException if called improperly
    */
   public Object authenticateRemote(String identity)
   throws RestrictedUseException
   {
      // restrict the use of this method
      checkCallerAbort("com.goldencode.p2j.net.RouterSessionManager.authenticateRemote");
      
      if (!cfg.isServer())
         return null;

      // Safely query cached data
      SecurityCache sc = getCache(); 

      // parse the identity
      String[] parts = identity.split("/");
      int len = parts.length;
      if (len < 1 || len > 2)
         return null;

      String procID = parts[0];
      String userID = (len == 2 ? parts[1] : null);

      //
      // Create new security context
      //

      SecurityContext newContext = createSecurityContext(sc, procID, userID);

      if (newContext == null || sc != getCache())
      {
         dTrace("authenticateRemote: newer security cache generation available; " +
                "authentication aborted.");
         return null;
      }

      //
      // Create new session
      //

      Session sess = new Session(identity, sc, newContext);
      synchronized (listOfSess)
      {
         listOfSess.add(sess);
      }
      assignContext(newContext);

      return newContext;
   }

   /**
    * Destroys the security context associated with the given key.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *   <li> com.goldencode.p2j.net.BaseSession.cleanupContext()
    *   <li> com.goldencode.p2j.net.Queue.stop()
    * </ul>
    *
    * @param    key
    *           The security context.
    *
    * @throws   RestrictedUseException
    *           If called improperly from unauthorized code.
    */
   public void terminateSession(Object key)
   throws RestrictedUseException
   {
      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.net.BaseSession.cleanupContext",
         "com.goldencode.p2j.net.Queue.stop"
      };
      checkCallerAbort(callers);
      
      if (!(key instanceof SecurityContext))
      {
         throw new IllegalStateException("Invalid security context key.");
      }
      
      Session sess = removeSession((SecurityContext) key);
      
      if (!cfg.isServer())
         return;

      // log the context deletion event
      SecurityContext ctx = sess.getSecurityContext();
      
      // TODO: shouldn't this use the LOGGER?
      if (ctx.isInitial())
         System.out.println("*** terminating Initial Security Context ***");
      
      // unassign context and do cleanup if it's time
      boolean finalUse = unassignContext(ctx);
      if (finalUse)
      {
         // ensure the context is valid for the lifespan of cleanup
         SecurityContextStack cts = SecurityContextStack.getContext();

         if (cts == null)
         {
            setInitialSecurityContextWorker(false, null);
            cts = SecurityContextStack.getContext();
         }

         // there are cases when there is already an additional context
         // attached... and this call will fail. so, we pop the previous 
         // context before doing the cleanup
         SecurityContext old = cts.pop();
         try
         {
            cts.push(ctx);
            ctx.cleanup();
            ctx = cts.pop();
         }
         finally
         {
            if (old != null)
               cts.push(old);
         }

         logGoingContext(ctx);
      }

   }
   
   /**
    * Destroys the security context associated with the given key.
    * <p>
    * <b>Restricted Use</b>. This method checks the caller to be 
    * <code>com.goldencode.p2j.main.ClientCore.start()</code>.
    *
    * @param    key
    *           The security context.
    *
    * @throws   RestrictedUseException
    *           If called improperly or from unauthorized code.
    */
   public void terminateSingle(Object key) 
   throws RestrictedUseException
   {
      if (!cfg.isServer())
         return;
      
      // restrict the use of this method
      checkCallerAbort("com.goldencode.p2j.main.ClientCore.start");
      
      if (!(key instanceof SecurityContext))
      {
         throw new IllegalStateException("Invalid security context key.");
      }
      
      SecurityContext context = (SecurityContext) key;
      
      // get rid of the session
      removeSession(context);
      
      // DON'T move context-locals back to the initial context! Instead the
      // next call will force cleanup of both the effective and initial
      // contexts.
      popAllSecurityContext();
   }

   /**
    * Implements client side standard authentication logic.
    * <p>
    * Returns a byte array to be transmitted to the server as authentication
    * input.  The userid and password values will be obtained by prompting
    * the user using <code>stdin</code> and <code>stdout</code>.
    * <p>
    * If this method is called, it simply provides a userID and password no
    * matter what authentication mode is. It works, because certificates are
    * verified before the call, if the authentication mode required that.
    * <p>
    * This logic allows specifying 
    * <code>com.goldencode.p2j.security.SecurityManager</code> as a hook name.
    *
    * @param    parameters
    *           Additional configuration parameters. Not used in this
    *           implementation.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes that is taken as an authentication input.
    *           The array is suitable for passing to {@link #serverAuthHook} 
    *           for authentication processing.
    */
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      return clientAuthHookWorker(parameters, code, cfg);
   }
   
   /**
    * Implements server side standard authorization logic.
    * <p>
    * Accepts the byte array produced by the client side authorization hook
    * as the authentication input, and custom parameters.
    *
    * @param    auth
    *           The authentication input from the client in a form that is
    *           created using {@link #packageIdPassword}.
    * @param    parameters
    *           Additional configuration parameters taken from the directory.
    *           Not used in this implementation.
    *
    * @return   The authentication result.
    */
   public AuthenticationResponse serverAuthHook(byte[] auth,
                                                String parameters)  
   {
      if (!cfg.isServer())
         return null;

      // parse the byte array input
      ByteArrayInputStream bis = new ByteArrayInputStream(auth);
      DataInputStream      dis = new DataInputStream(bis);
      
      String userId = null;
      String pw     = null;
      char[] userPw = null;
      int    pwsize = 0;
      
      try
      {
         userId = dis.readUTF();
         pw     = dis.readUTF();
         pwsize = pw.length();
         userPw = pw.toCharArray();
      }
      
      catch (IOException ioe)
      {
         return new AuthenticationResponse(null,
                                           AUTH_RESULT_UNSPECIFIED_FAILURE);
      }
      
      // WARNING: the security cache accessed here may be different than the caller's security
      // cache generation
      
      // Safely query cached data
      SecurityCache sc = getCache(); 

      // locate the account and its password hash
      dTrace("Received user ID <" + userId + ">");
      Account acc = sc.getAccountById(userId);
      
      if (acc == null)
      {
         dTrace("No account for ID <" + userId + ">");
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }
      
      // TODO: Why is this restriction needed? By doing this, one cannot
      //       use the custom auth method to authenticate processes, right?
      if (acc.getAccountType() != Account.ACC_USER)
      {
         dTrace("Wrong account type for ID <" + userId + ">");
         return new AuthenticationResponse(null,
                                           AUTH_RESULT_UNSPECIFIED_FAILURE);
      }
      
      UserAccount user = (UserAccount) acc;
                              
      if (user.getAuthMode() == AUTH_MODE_X509)
      {
         dTrace("Account <" + userId +
                "> uses x509 auth mode, cancelling custom or id/pw auth");
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }
      
      // verify password
      if (debugLevel.intValue() <= DATA.intValue())
         dData("Received password <" + pw + ">");
      else
         dTrace("Received password");
      
      if (checkPassword(new String(userPw), user))
      {
         agePassword(user);
         return new AuthenticationResponse(userId, AUTH_RESULT_SUCCESS);
      }

      return new AuthenticationResponse(null, AUTH_RESULT_INVALID_PASSWORD);
   }

   /**
    * Initializes security context switching for the calling thread.
    * This call is supposed to be repeatable for permanent threads to allow
    * proper security cache refresh function.
    * <p>
    * This call immediately exits if called from the client.
    * <p>
    * If this call is made repeatedly, it checks whether the security cache is
    * refreshed compared to the generation of the cache used to create the
    * initial security context currently in use. If this is the case, the old
    * security context is released and replaced with the current one.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> com.goldencode.p2j.net.RouterSessionManager.setInitialContext()
    *  <li> com.goldencode.p2j.net.StandardServer.bootstrap()
    *  <li> com.goldencode.p2j.net.StandardServer$1.run()                   
    * </ul>
    *
    * @throws RestrictedUseException if called improperly
    */
   public void setInitialSecurityContext()
   throws RestrictedUseException
   {
      // optimally, this would be done after the checkCaller below, but that
      // call is very expensive and by putting this here we allow the call
      // to this method but the caller gets no service so security is not
      // really reduced
      if (!cfg.isServer())
         return;

      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.net.RouterSessionManager.setInitialContext",
         "com.goldencode.p2j.main.StandardServer.bootstrap",
         "com.goldencode.p2j.main.StandardServer$1.run"
      };
      checkCallerAbort(callers);

      setInitialSecurityContextWorker(false, null);
   }
   
   /**
    * Initializes security context switching for the calling thread making
    * sure that the context is newly created and is not shared with any other
    * thread. The context will not be associated with any specific account
    * without a subsequent authentication. This is different than standard
    * server threads which have an initial context that is shared with other
    * server threads AND which are naturally associated with the server's
    * account (and thus with the server's rights).
    * <p>
    * This call is supposed to be repeatable for permanent threads to allow
    * proper security cache refresh function.
    * <p>
    * This call immediately exits if called from the client.
    * <p>
    * If this call is made repeatedly, it checks whether the security cache is
    * refreshed compared to the generation of the cache used to create the
    * initial security context currently in use. If this is the case, the old
    * security context is released and replaced with the current one.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> com.goldencode.p2j.main.ServerDriver$SingleProcessClient.run()  
    * </ul>
    *
    * @throws RestrictedUseException if called improperly
    */
   public void setUniqueInitialSecurityContext()
   throws RestrictedUseException
   {
      // optimally, this would be done after the checkCaller below, but that
      // call is very expensive and by putting this here we allow the call
      // to this method but the caller gets no service so security is not
      // really reduced
      if (!cfg.isServer())
         return;

      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.main.ServerDriver$SingleProcessClient.run"
      };
      checkCallerAbort(callers);

      setInitialSecurityContextWorker(true, null);
   }
   
   /**
    * Initializes the security context switching for the calling thread using
    * the given context stack as the source of the initial and effective
    * contexts.
    *
    * @param    stack
    *           Context stack to use instead of creating a new one. This
    *           allows the caller to override the initial and effective
    *           contexts.
    */
   void forceSecurityContext(SecurityContextStack stack)
   {
      setInitialSecurityContextWorker(false, stack);
   }
   
   /**
    * Remove the session (which is associated with the given context) from the
    * master session list.
    *
    * @param    key
    *           The security context.
    *
    * @return   The session that was removed.
    */
   private Session removeSession(SecurityContext key)
   {
      // locate the session
      Session sess = locateSession(key);

      // purge the session data
      synchronized (listOfSess)
      {
         listOfSess.remove(sess);
      }
      
      return sess;
   }

   /**
    * Implements client side standard authentication logic.
    * <p>
    * Returns a byte array to be transmitted to the server as authentication
    * input.  If not contained in the given bootstrap configuration, the userid
    * and password values will be obtained by prompting the user using
    * <code>stdin</code> and <code>stdout</code>.
    * <p>
    * If this method is called, it simply provides a userID and password no
    * matter what authentication mode is. It works, because certificates are
    * verified before the call, if the authentication mode required that.
    *
    * @param    parameters
    *           Additional configuration parameters. Not used in this
    *           implementation.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    * @param    config
    *           Configuration data upon which to base default processing.
    *
    * @return   Array of bytes that is taken as an authentication input.
    *           The array is suitable for passing to {@link #serverAuthHook} 
    *           for authentication processing.
    */
   private byte[] clientAuthHookWorker(Map<String, Object> parameters,
                                       int code,
                                       BootstrapConfig config)
   {
      if (config.isServer())
         return null;
      
      String id = null;
      String pw = null;
      boolean stored = true;
      
      try
      {
         id = config.getString("access", "subject", "id", null);
         
         if (id == null)
         {
            id = new String(Utils.prompt("User ID  :"));
            stored = false;
         }
         
         pw = config.getString("access", "password", "user", null);
         
         if (pw == null)
         {
            pw = new String(Utils.prompt("Password :"));
            stored = false;
         }
         
         config.setConfigItem("access", "password", "user", "password");
      }
      
      catch (Exception exc)
      {
         // pass through any failures
      }
      
      // if there was a failure and this input comes from the config, fail 
      if (stored && code > AUTH_RESULT_SUCCESS)
      {
         return null;
      }
      
      return packageIdPassword(id, pw);
   }
   
   /**
    * Drops any current context (if any) and establishes an initial context
    * for the thread.
    *
    * @param    unique
    *           <code>true</code> to instantiate a new context that is not
    *           associated with the current server account and does not share
    *           any context-local data. <code>false</code> to get the standard
    *           server-wide initial context which is associated with the
    *           server's account AND shares context-local data with all other
    *           server threads that also share this initial context. This
    *           parameter is only honored if the <code>context</code> parameter
    *           is <code>null</code>.
    * @param    stack
    *           Optional context stack to use instead of creating a new one.
    *           This allows the caller to override the initial and effective
    *           contexts.
    */
   private void setInitialSecurityContextWorker(boolean              unique,
                                                SecurityContextStack stack)
   {
      int oldSerial = -1;
      
      // access the current context stack
      SecurityContextStack cts = SecurityContextStack.getContext();
      
      // if there is currently a context associated, get the serial number
      if (cts != null)
      {
         oldSerial = cts.getCache().getSerial();
      }

      // safely query the current cache
      SecurityCache sc = getCache(); 

      // (re)establish initial security context (this occurs when there is
      // no context or when the context was defined in an older generation
      // of the cache)
      if (sc.getSerial() != oldSerial)
      {
         if (oldSerial != -1)
            dropInitialSecurityContext();
         
         SecurityContext initial   = null;
         SecurityContext effective = null;
         
         if (stack == null)
         {
            initial = unique ? sc.getUniqueInitialSecurityContext()
                             : sc.getInitialSecurityContext();
                             
            stack = new SecurityContextStack(initial);
         }
         else
         {
            initial   = stack.getInitial();
            effective = stack.getEffectiveContext(); 
         }
         
         // sets the thread-local pointer to the stack
         SecurityContextStack.setContext(stack);
         
         // assigns the initial context
         logContext(sc, initial);
         assignContext(initial);
         
         // update reference count etc for the effective context if needed
         if (effective != null && !effective.isInitial())
         {
            logContext(sc, effective);
            assignContext(effective);
         }
         
         DirectoryService ds = DirectoryService.getInstance();
         ds.bind();
      }
   }

   /**
    * Releases the initial security context assigned to the calling thread.
    * The initial security context can only be released if it is currently
    * assigned and is the only context in the stack.
    * <p>
    * If called improperly, has no effect.
    */
   public void dropInitialSecurityContext()
   {
      if (!cfg.isServer())
         return;

      // access the current context stack
      SecurityContextStack cts = SecurityContextStack.getContext();
      if (cts == null)
         return;

      SecurityContext ctx = cts.getEffectiveContext();
      if (!ctx.isInitial())
         return;

      // safely query cached data and log the context switch event
      SecurityCache sc = getCache(); 
      logContext(sc, null);

      // drop the initial security context
      SecurityContextStack.setContext(null);
      unassignContext(ctx);
      
      // remove the thread ID from the map
      synchronized (threadIDs)
      {
         threadIDs.remove(Thread.currentThread()); 
      }
   }

   /**
    * Temporarily switches to a user context to serve the associated request.  
    * The calling thread has to have the initial security context.
    * <p>
    * The use count for the target security context is incremented by 1.
    *
    * @param    key
    *           security context key
    *
    * @throws   RestrictedUseException
    *           If called improperly.
    */
   public void pushAndSwitchSecurityContext(Object key)
   throws RestrictedUseException
   {
      if (!cfg.isServer())
         return;
      
      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.net.BaseSession.setContext"
      };
      checkCallerAbort(callers);

      pushContextWorker(key);
   }
      
   /**
    * Terminates all security contexts associated with the current thread,
    * including the initial context. This MUST be used when terminating
    * contexts obtained with {@link #setUniqueInitialSecurityContext}. If this
    * is not used, then the unique initial contexts will never be cleaned up
    * and there will be a memory leak.
    * <p>
    * The use count for the all security contexts is decremented by 1.  
    * If it becomes 0, it gets deleted immediately.
    *
    * @throws   RestrictedUseException
    *           If called from an unauthorized location.
    */
   void popAllSecurityContext()
   throws RestrictedUseException
   {
      if (!cfg.isServer())
         return;

      SecurityContextStack cts = SecurityContextStack.getContext();
      
      if (cts == null)
         throw new RestrictedUseException("no security context established");
      
      SecurityContext ctx = cts.getEffectiveContext();
      
      if (ctx == null)
         throw new RestrictedUseException("illegal context switch");
      
      // exit any non-initial context using the normal worker
      if (!ctx.isInitial())
      {
         popContextWorker();
         ctx = cts.getInitial();
      }
      
      // exit the initial context
      endContext(cts, ctx);
   }
   
   /**
    * Restores the initial security context of the calling thread when it is
    * done with the user request.  The calling thread has to have its security
    * context switched to a user context prior to this call.
    * <p>
    * The use count for the current security context is decremented by 1.  
    * If it becomes 0, it gets deleted immediately.
    *
    * @throws   RestrictedUseException
    *           If called from an unauthorized location.
    */
   public void popAndRestoreSecurityContext()
   throws RestrictedUseException
   {
      if (!cfg.isServer())
         return;

      // restrict the use of this method
      String[] callers = new String[]
      {
         "com.goldencode.p2j.net.RouterSessionManager.restoreContext"
      };
      checkCallerAbort(callers);
      
      popContextWorker();
   }
   
   /**
    * Searches abstract resource registry for a plugin that is responsible
    * for the resources of the specified type.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread, since in the current cache the resource plugin can be
    * either registered under different ID or even be dropped.
    * 
    * @param resourceTypeName  
    *        resource type name
    *
    * @return <code>AbstractResource</code> object
    */
   public AbstractResource getPluginInstance(String resourceTypeName)
   {
      if (!cfg.isServer())
         return null;

      // Pick up the generation of the security cache
      SecurityContextStack scs = SecurityContextStack.getContext();
      if (scs == null)
         return null;
      SecurityCache sc = scs.getCache(); 

      int sysOrdinal = sc.getRegistryOrdByTypeName(resourceTypeName);
      if (sysOrdinal == -1)
         return null;
      else
         return sc.getRegistryByOrd(sysOrdinal).getPlugin();
   }

   /**
    * Searches the cache for a previously made decision. 
    * <p>
    * Plugins call this method to check whether a cached access check decision
    * is available.  Cached access check decisions are identified by triplets 
    * of {resourceId, instanceName, mode} exactly as they were specified when 
    * the decision was taken.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  resourceId
    *         resource identity
    *
    * @param  instanceName
    *         resource instance name
    *
    * @param  mode
    *         access mode (requested rights) the decision was made about
    * @return <code>Boolean</code> that wraps a decision, or <code>null</code> 
    *         if not available. 
    */
   public Boolean getCachedDecision(int resourceId, String instanceName, 
                                    int mode)
   {
      if (!cfg.isServer())
         return null;

      SecurityContextStack ctx = SecurityContextStack.getContext();
      Map[] cache = ctx.getDecisionCache();

      // Pick up the generation of the security cache
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();

      if (resourceId < 0 || resourceId >= nres)
         return null;

      Boolean found = Decision.searchDecisions(cache[resourceId], 
                                               instanceName, mode);
      if (found != null)
      {
         // for audit log, use always the current security cache
         sc = getCache();
         logResource(sc, resourceId, instanceName, mode, found.booleanValue(),
                     "cached");
      }

      return found;
   }
   
   /**
    * Creates a search handle.
    * <p>
    * Plugins call this method to initiate the access rights search for 
    * the specified resource instance and requested rights.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  resourceId
    *         resource identity
    *
    * @param  instanceName
    *         resource instance name
    *
    * @param  mode
    *         access mode (requested rights) the decision was made about
    * @return integer handle value for the new search 
    */
   public int openRightsSearch(int resourceId, String instanceName, 
                               int mode)   
   {
      if (!cfg.isServer())
         return 0;

      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();

      if (resourceId < 0 || resourceId >= nres)
      {
         return 0;
      }

      ResourceRegistry reg = sc.getRegistryByOrd(resourceId);
      Map searches = reg.getOpenSearches();
      AccessControlList[] acls = reg.getACLs();

      int[] checkList = ctx.getCheckList();

      // create new open search
      Search s = new Search(instanceName, mode, checkList, acls);
      int handle = s.getHandle();

      // put it into the map of open searches
      synchronized (searches)
      {
         searches.put(new Integer(handle), s);
      }

      // clear preexisting cached decision if any
      Map[] cache = ctx.getDecisionCache();
      Decision.clearDecisions(cache[resourceId], instanceName, mode);

      return handle;

   }

   /**
    * Returns an instance of Rights interface to be used next in the access 
    * rights check loop.
    * <p>
    * Plugins cast Rights to their own classes and use their custom methods 
    * to perform the check. 
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  handle
    *         search handle 
    *
    * @return next available instance of <code>Rights</code> 
    */
   public Rights getNextRights(int handle)
   {
      if (!cfg.isServer())
         return null;

      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 

      int nres = sc.getNumResources();
      ResourceRegistry reg = null;
      Search s = null;
      Object key = new Integer(handle);
      int i = 0;

      // find the search by its handle
      for (i = 0; i < nres; i++)
      {
         reg = sc.getRegistryByOrd(i);
         Map searches = reg.getOpenSearches();
         synchronized (searches)
         {
            if (searches.containsKey(key))
            {
               s = (Search)searches.get(key);
               break;
            }
         }
      }
      if (i == nres)
      {
         // handle not found
         return null;
      }
      
      // query the search for next Rights
      return s.next();
   }
   
   /**
    * Terminates the open search of Rights and frees up the related resources.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  handle
    *         search handle 
    *
    * @param  decision
    *         <code>true</code> or <code>false</code> decision just made 
    *
    * @param  cache
    *         <code>true</code> if the decision has to be cached
    */
   public void closeRightsSearch(int handle, boolean decision, boolean cache)
   {
      if (!cfg.isServer())
         return;

      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 

      int nres = sc.getNumResources();
      ResourceRegistry reg = null;
      Search s = null;
      Object key = new Integer(handle);
      int i = 0;

      // find the search by its handle
      for (i = 0; i < nres; i++)
      {
         reg = sc.getRegistryByOrd(i);
         Map searches = reg.getOpenSearches();
         synchronized (searches)
         {
            if (searches.containsKey(key))
            {
               s = (Search) searches.get(key);
               // dispose of the search
               searches.remove(key);
               break;
            }
         }
      }
      if (i == nres)
      {
         // handle not found
         return;
      }

      // log the access decision event
      Map[] cmap = ctx.getDecisionCache();

      // for audit log, use always the current security cache
      sc = getCache();

      // cache or not to cache?
      if (!cache)
      {
         logResource(sc, i, s.getInstanceName(), s.getMode(), decision);
         return;
      }
      
      logResource(sc, i, s.getInstanceName(), s.getMode(), decision, "cache");
      Decision.cacheDecision(cmap[i], s.getInstanceName(), s.getMode(),
                             decision);
   }
   
   /**
    * Creates a search handle. This search handle is a temporary search, 
    * created within the context of the currently authenticated subject. This 
    * is used for cases when the current user needs to determine the actual
    * permissions for another subject, different then the authenticated one.
    * <p>
    * Plugins call this method to initiate the access rights search for 
    * the specified resource instance and requested rights.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    * <p>
    * The search handle is created for the specified subject.
    * 
    * @param  subject
    *         subject ID to which this search belongs.
    * 
    * @param  resourceId
    *         resource identity
    *
    * @param  instanceName
    *         resource instance name
    *
    * @param  mode
    *         access mode (requested rights) the decision was made about
    * @return integer handle value for the new search 
    */
   public int openRightsSearch(String subject, 
                               int resourceId, 
                               String instanceName, 
                               int mode)   
   {
      if (!cfg.isServer() || subject == null)
         return 0;
      
      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();
      
      if (resourceId < 0 || resourceId >= nres)
      {
         return 0;
      }
      
      ResourceRegistry reg = sc.getRegistryByOrd(resourceId);
      Map searches = reg.getOpenSearches(subject, true);
      AccessControlList[] acls = reg.getACLs();
      
      // the checklist contains only the ordinal for the specified subject;
      // this means the permission search will be performed only among the
      // ACLs defined explicitly for this subject; if the subject is an user
      // which is part of some group, the group permissions will *NOT* 
      // be searched
      int[] checkList = new int[] { getCache().getOrdinalById(subject) };
      
      // create new open search
      Search s = new Search(instanceName, mode, checkList, acls);
      int handle = s.getHandle();
      
      synchronized (searches)
      {
         // put it into the map of open searches
         searches.put(new Integer(handle), s);
      }
      
      // don't care about caching, just return the handle
      return handle;
   }
   
   /**
    * Returns an instance of Rights interface to be used next in the access 
    * rights check loop. This search handle is a temporary search, 
    * created within the context of the currently authenticated subject. This 
    * is used for cases when the current user needs to determine the actual
    * permissions for another subject, different then the authenticated one.
    * <p>
    * Plugins cast Rights to their own classes and use their custom methods 
    * to perform the check. 
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  subject
    *         subject ID to which the search handle belongs.
    *  
    * @param  handle
    *         search handle
    *
    * @return next available instance of <code>Rights</code> 
    */
   public Rights getNextRights(String subject, int handle)
   {
      if (!cfg.isServer() || subject == null)
         return null;
      
      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      
      int nres = sc.getNumResources();
      ResourceRegistry reg = null;
      Search s = null;
      Object key = new Integer(handle);
      int i = 0;
      
      // find the search by its handle
      for (i = 0; i < nres; i++)
      {
         reg = sc.getRegistryByOrd(i);
         Map searches = reg.getOpenSearches(subject, false);
         if (searches != null)
         {
            synchronized (searches)
            {
               if (searches.containsKey(key))
               {
                  s = (Search)searches.get(key);
                  break;
               }
            }
         }
      }
      if (i == nres)
      {
         // handle not found
         return null;
      }
      
      // query the search for next Rights
      return s.next();
   }
   
   /**
    * Terminates the open search of Rights and frees up the related resources.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  subject
    *         subject ID to which the search handle belongs.
    * 
    * @param  handle
    *         search handle 
    *
    * @param  decision
    *         <code>true</code> or <code>false</code> decision just made 
    */
   public void closeRightsSearch(String subject, int handle, boolean decision)
   {
      if (!cfg.isServer() || subject == null)
         return;
      
      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      
      int nres = sc.getNumResources();
      ResourceRegistry reg = null;
      Search s = null;
      Object key = new Integer(handle);
      int i = 0;
      
      // find the search by its handle
      for (i = 0; i < nres; i++)
      {
         reg = sc.getRegistryByOrd(i);
         Map searches = reg.getOpenSearches(subject, false);
         if (searches != null)
         {
            synchronized (searches)
            {
               if (searches.containsKey(key))
               {
                  // dispose of the search
                  s = (Search)searches.get(key);
                  searches.remove(key); 
                  
                  if (searches.isEmpty())
                  {
                     reg.removeOpenSearches(subject);
                  }
                  
                  break;
               }
            }
         }
      }
      if (i == nres)
      {
         // handle not found
         return;
      }
      
      logResource(sc, i, s.getInstanceName(), s.getMode(), decision);
   }

   /**
    * Compiles and evalautes an expression. 
    * <p>
    * The expression may refer to Security Manager's variables and functions
    * and to the plugin's variables and functions, should one decide to expose
    * any.
    * <p>
    * The current implementation arbitrates names clashes using natural 
    * scoping provided by Expression Engine, which can be modified by using
    * qualifiers in the expression.
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    *
    * @param  resourceId
    *         resource identity
    *
    * @param  link
    *         an Object that supplies the raw data for this computation and is
    *         shared between the resource plugin and its library of exports
    *
    * @param  expr
    *         expression to compile and evaluate 
    *
    * @return the result. If expression compilation fails, 
    *         <code>null</code> is returned
    */
   public synchronized Object evaluate(int resourceId, Object link, 
                                       String expr)   
   {
      if (!cfg.isServer())
         return null;

      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();

      if (resourceId < 0 || resourceId >= nres)
      {
         return null;
      }

      // linking the resource plugin and rights with the expression
      // TODO: the current generation of the security cache used here may be different than the
      // one used when resolving the expression; need to explicitly pass the security cache 
      // generation saved in "sc" to the resolver
      AbstractResource ar = sc.getRegistryByOrd(resourceId).getPlugin();
      Resolver resolver = sc.getResolver();

      ar.associate(link);
      resolver.setCurrentScope(ar);

      // compilation and evaluation
      Object result = null;
      try
      {
         Expression express = new Expression(resolver, expr);
         result = express.execute();
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
         result = null;
      }

      ar.disassociate();

      return result;
   }

   /**
    * Notifies the Security Manager about a batch editing session to be 
    * started for the specified directory branch.
    * If the branch is either /security or any of its branches, notes the
    * state of batch editing.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> method openBatch in com.goldencode.p2j.directory.DirectoryService
    * </ul>
    * <p>
    * DirectoryService calls this method before it opens a batch. This method
    * checks the subject's access rights to the "system", "change" instance, 
    * if the batch is for any branch of "/security".
    * <p>
    * The action is different for master servers versus regular ones.
    * Only master servers are allowed to edit the P2J security directory.  
    * Regular servers cannot edit, but still need this call as a way of 
    * refreshing their internal caches. 
    * <p>
    * Note that the security cache has to be taken from the security context 
    * of the thread.
    * 
    * @param nodeId
    *           Subtree which will be edited.
    * @return One of the following codes:
    *         <ul>
    *           <li>-1 if the subject has no rights to open batch or another
    *                  batch is currently open
    *           <li> 0 if the batch can be opened
    *           <li> 1 if the batch can be opened R/O; not a master server
    *         </ul> 
    */
   public int openBatch(String nodeId)
   {
      // optimally, this would be done after the checkCaller below, but that
      // call is very expensive and by putting this here we allow the call
      // to this method but the caller gets no service so security is not
      // really reduced
      if (!cfg.isServer())
         return -1;
      
      // restrict the use of this method
      if (!checkCaller(
          "com.goldencode.p2j.directory.DirectoryService.openBatch"))
         return -1;

      // accessing security context
      if (!init)
      {
         return 0;
      }
      
      SecurityContextStack ctx = SecurityContextStack.getContext();
      if (ctx.isEditing())
         return -1;

      // check to see if the batch is under /security
      if (!nodeId.equals("/security") && !nodeId.startsWith("/security/"))
         return 0;
      
      // fail the call if not sufficient access rights (if appl context)
      SecurityContext sctx = ctx.getEffectiveContext();
      if (!sctx.isInitial())
      {
         SystemResource sr = (SystemResource)getPluginInstance("system");
         if (!sr.checkChangeAccess())
            return -1;
      }

      // note editing batch in the context
      ctx.setEditing(true);

      // Pick up the generation of the security cache
      SecurityCache sc = ctx.getCache(); 

      // do not open R/W security editing batch for non-masters
      if (!sc.getServerAccount().isMaster())
         return 1;
      else
         return 0;
   }

   /**
    * Checks to see if a batch editing is in progress.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> method isEditing in com.goldencode.p2j.directory.DirectoryService
    * </ul>
    * 
    * @return <code>true</code> if there is an open editing batch
    */
   public boolean isEditing()
   {
      // optimally, this would be done after the checkCaller below, but that
      // call is very expensive and by putting this here we allow the call
      // to this method but the caller gets no service so security is not
      // really reduced
      if (!cfg.isServer())
         return false;
      
      // restrict the use of this method
      if (!checkCaller(
          "com.goldencode.p2j.directory.DirectoryService.isEditing"))
         return false;

      // accessing security context
      return SecurityContextStack.getContext().isEditing();
   }

   /**
    * Notifies the Security Manager about a batch editing session to be
    * closed assuming the security cache to be refreshed.
    * <p>
    * DirectoryService calls this method after having successfully commited
    * the changes to the directory (disposition is <code>true</code>) or to
    * signal rollback.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> method closeBatchInt in
    *       com.goldencode.p2j.directory.DirectoryService
    * </ul>
    * 
    * @param    disposition
    *           If this parameter is <code>true</code> then changes need to
    *           be committed. Otherwise they are just thrown away and lock is
    *           removed.
    * @return   <code>true</code> if operation was successful and tree was
    *           successfully updated.
    */
   public boolean closeBatch(boolean disposition)
   {
      return closeBatch(disposition, true);
   }

   /**
    * Notifies the Security Manager about a batch editing session to be
    * closed.
    * <p>
    * DirectoryService calls this method after having successfully commited
    * the changes to the directory (disposition is <code>true</code>) or to
    * signal rollback.
    * <p>
    * <i><u>Restricted use</u></i>. This method checks the caller to be: 
    * <ul>
    *  <li> method closeBatchInt in
    *       com.goldencode.p2j.directory.DirectoryService
    * </ul>
    * 
    * @param    disposition
    *           If this parameter is <code>true</code> then changes need to
    *           be committed. Otherwise they are just thrown away and lock is
    *           removed.
    * @param    refresh
    *           when committing the batch, tells whether to refresh the
    *           security cache or not
    * @return   <code>true</code> if operation was successful and tree was
    *           successfully updated.
    */
   public boolean closeBatch(boolean disposition, boolean refresh)
   {
      // optimally, this would be done after the checkCaller below, but that
      // call is very expensive and by putting this here we allow the call
      // to this method but the caller gets no service so security is not
      // really reduced
      if (!cfg.isServer())
         return false;
      
      // restrict the use of this method
      if (!checkCaller(
          "com.goldencode.p2j.directory.DirectoryService.closeBatchInt"))
         return false;

      // accessing security context
      if (!init)
      {
         return true;
      }
      
      SecurityContextStack ctx = SecurityContextStack.getContext();
      if (!ctx.isEditing())
         return false;

      ctx.setEditing(false);

      // refreshing the security cache: no refresh if rollback or no request
      if (!disposition || !refresh)
         return true;

      // refreshing the security cache: no refresh if by a worker thread
      SecurityContext sctx = ctx.getEffectiveContext();
      if (sctx.isInitial())
         return true;

      // refreshing the security cache
      SecurityCacheRefresh ref = new SecurityCacheRefresh(SecurityCache.
                                                          LM_CONSOLE, null);
      Thread thread;
      thread = new Thread(secThreads, ref, "refresh");

      thread.start();

      return true;
   }

   /**
    * Checks access rights of the current subject with regards to the 
    * "shutdown" instance of the "system" abstract resource.
    *
    * @return <code>true</code> if access is allowed.
    */
   public boolean checkShutdownAccess()
   {
      if (!cfg.isServer())
         return false;

      SystemResource sr = (SystemResource)getPluginInstance("system");
      return sr.checkShutdownAccess();
   }

   /**
    * Conditionally outputs a debug message with an optional stack trace.
    *
    * @param  level
    *         Debug level of this message.
    * @param  message
    *         Text to be displayed.
    * @param  exc
    *         Exception containing the stack trace to be logged.  If
    *         <code>null</code>, only <code>message</code> will be logged.
    */
   public void debug(Level level, String message, Exception exc)
   {
      if (level.intValue() >= debugLevel.intValue())
      {
         if (exc != null)
         {
            LOG.logp(level,
                     "SecurityManager",
                     "",
                     LogHelper.generate(message),
                     exc);
         }
         else
         {
            LOG.logp(level,
                     "SecurityManager",
                     "",
                     LogHelper.generate(message));
         }
      }
   }

   /**
    * Conditionally outputs a debug message.
    *
    * @param  level
    *         Debug level of this message
    * @param  message
    *         Text to be displayed
    */
   public void debug(Level level, String message)
   {
      debug(level, message, null);
   }

   /**
    * Logs error messages.
    *
    * @param  message
    *         text to be displayed
    */
   public void dErr(String message)
   {
      debug(Level.SEVERE, message);
   }

   /**
    * Logs warnings.
    *
    * @param  message
    *         text to be displayed
    */
   public void dWarn(String message)
   {
      debug(Level.WARNING, message);
   }

   /**
    * Logs statistics.
    *
    * @param  message
    *         text to be displayed
    */
   public void dStat(String message)
   {
      debug(Level.INFO, message);
   }

   /**
    * Logs detailed lists.
    *
    * @param  message
    *         text to be displayed
    */
   public void dList(String message)
   {
      debug(Level.FINE, message);
   }

   /**
    * Logs detailed traces.
    *
    * @param  message
    *         text to be displayed
    */
   public void dTrace(String message)
   {
      debug(Level.FINER, message);
   }

   /**
    * Logs detailed data, possibly including sensitive information.
    *
    * @param  message
    *         text to be displayed
    */
   public void dData(String message)
   {
      debug(Level.FINEST, message);
   }

   /**
    * Returns <code>true</code> if this is the server.
    *
    * @return   <code>true</code> if on the server.
    */
   public boolean isServer()
   {
      return cfg.isServer();
   }

   /**
    * Returns <code>true</code> if this is a client.
    *
    * @return   <code>true</code> if on the client.
    */
   public boolean isClient()
   {
      return !isServer();
   }

   /**
    * Gets the named piece of account extension data of type int for the
    * current user account.
    *
    * @param extName
    *        account extension name
    *
    * @return extension value wrapped into <code>Integer</code> or
    *         <code>null</code> if no such extension exists.
    */
   public Integer getExtInteger(String extName)
   {
      return getExtInteger(extName, getUserId());
   }

   /**
    * Gets the named piece of account extension data of type boolean for the
    * current user account.
    *
    * @param extName
    *        account extension name
    *
    * @return extension value wrapped into <code>Boolean</code> or
    *         <code>null</code> if no such extension exists.
    */
   public Boolean getExtBoolean(String extName)
   {
      return getExtBoolean(extName, getUserId());
   }

   /**
    * Gets the named piece of account extension data of type String for the
    * current user account.
    *
    * @param extName
    *        account extension name
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public String getExtString(String extName)
   {
      return getExtString(extName, getUserId());
   }

   /**
    * Gets the named piece of account extension data of type String for the
    * current user account.
    *
    * @param extName
    *        account extension name
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public DateValue getExtDate(String extName)
   {
      return getExtDate(extName, getUserId());
   }

   /**
    * Gets the named piece of account extension data of type byte[] for the
    * current user account.
    *
    * @param extName
    *        account extension name
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public byte[] getExtBytes(String extName)
   {
      return getExtBytes(extName, getUserId());
   }

   /**
    * Gets the named piece of account extension data of type int for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    *
    * @return extension value wrapped into <code>Integer</code> or
    *         <code>null</code> if no such extension exists.
    */
   public Integer getExtInteger(String extName, String subjectId)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return null;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return null;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // get the target directory object's "value" attribute of type Integer
      Integer value = ds.getNodeInteger(new String(path + "/" + extName),
                                       "value");
      
      // deactivate directory services
      ds.unbind();

      return value;
   }

   /**
    * Gets the named piece of account extension data of type boolean for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    *
    * @return extension value wrapped into <code>Boolean</code> or
    *         <code>null</code> if no such extension exists.
    */
   public Boolean getExtBoolean(String extName, String subjectId)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return null;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return null;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // get the target directory object's "value" attribute of type Integer
      Boolean value = ds.getNodeBoolean(new String(path + "/" + extName),
                                        "value");
      
      // deactivate directory services
      ds.unbind();

      return value;
   }

   /**
    * Gets the named piece of account extension data of type String for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public String getExtString(String extName, String subjectId)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return null;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return null;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // get the target directory object's "value" attribute of type Integer
      String value = ds.getNodeString(new String(path + "/" + extName),
                                      "value");
      
      // deactivate directory services
      ds.unbind();

      return value;
   }

   /**
    * Gets the named piece of account extension data of type date for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public DateValue getExtDate(String extName, String subjectId)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return null;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return null;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // get the target directory object's "value" attribute of type Integer
      DateValue value = ds.getNodeDate(new String(path + "/" + extName),
                                      "value");
      
      // deactivate directory services
      ds.unbind();

      return value;
   }

   /**
    * Gets the named piece of account extension data of type byte[] for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    *
    * @return extension value or
    *         <code>null</code> if no such extension exists.
    */
   public byte[] getExtBytes(String extName, String subjectId)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return null;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return null;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // get the target directory object's "value" attribute of type Integer
      byte[] value = ds.getNodeByteArray(new String(path + "/" + extName),
                                         "value");
      
      // deactivate directory services
      ds.unbind();

      return value;
   }

   /**
    * Sets the named piece of account extension data of type int for the
    * current user account.
    *
    * @param extName
    *        account extension name
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtInteger(String extName, int value)
   {
      return setExtInteger(extName, getUserId(), value);
   }

   /**
    * Sets the named piece of account extension data of type boolean for the
    * current user account.
    *
    * @param extName
    *        account extension name
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtBoolean(String extName, boolean value)
   {
      return setExtBoolean(extName, getUserId(), value);
   }

   /**
    * Sets the named piece of account extension data of type String for the
    * current user account.
    *
    * @param extName
    *        account extension name
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtString(String extName, String value)
   {
      return setExtString(extName, getUserId(), value);
   }

   /**
    * Sets the named piece of account extension data of type byte[] for the
    * current user account.
    *
    * @param extName
    *        account extension name
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtBytes(String extName, byte[] value)
   {
      return setExtBytes(extName, getUserId(), value);
   }

   /**
    * Sets the named piece of account extension data of type int for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtInteger(String extName, String subjectId, int value)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return false;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return false;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return false;
      if (!ds.bind())
         return false;

      // open an editing batch
      if (!ds.openBatch(new String(path)))
      {
         ds.unbind();
         return false;
      }

      // set the target directory object's "value" attribute of type Integer
      boolean rc = ds.setNodeInteger(new String(path + "/" + extName),
                                     "value", value);

      // close the batch
      boolean brc = ds.closeBatch(rc);

      // deactivate directory services
      ds.unbind();

      if (!brc)
         return false;

      return rc;
   }

   /**
    * Sets the named piece of account extension data of type boolean for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtBoolean(String extName, String subjectId, 
                                boolean value)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return false;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return false;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return false;
      if (!ds.bind())
         return false;

      // open an editing batch
      if (!ds.openBatch(new String(path)))
      {
         ds.unbind();
         return false;
      }

      // set the target directory object's "value" attribute of type Integer
      boolean rc = ds.setNodeBoolean(new String(path + "/" + extName),
                                     "value", value);

      // close the batch
      if (!ds.closeBatch(true))
      {
         ds.unbind();
         return false;
      }

      // deactivate directory services
      ds.unbind();

      return rc;
   }

   /**
    * Sets the named piece of account extension data of type String for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtString(String extName, String subjectId, String value)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return false;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return false;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return false;
      if (!ds.bind())
         return false;

      // open an editing batch
      if (!ds.openBatch(new String(path)))
      {
         ds.unbind();
         return false;
      }

      // set the target directory object's "value" attribute of type Integer
      boolean rc = ds.setNodeString(new String(path + "/" + extName),
                                    "value", value);

      // close the batch
      if (!ds.closeBatch(true))
      {
         ds.unbind();
         return false;
      }

      // deactivate directory services
      ds.unbind();

      return rc;
   }

   /**
    * Sets the named piece of account extension data of type byte[] for the
    * specified account.
    *
    * @param extName
    *        account extension name
    * @param subjectId
    *        subject ID
    * @param value
    *        new value for the account extension
    *
    * @return <code>true</code> if successfully set the value
    */
   public boolean setExtBytes(String extName, String subjectId, byte[] value)
   {
      // check the access rights
      if (getUserId().compareToIgnoreCase(subjectId) != 0)
      {  
         SystemResource sr = (SystemResource)getPluginInstance("system");
         boolean decision = sr.checkExtensionsAccess();
         if (!decision)
            return false;
      }

      String path = locateExtensionPath(subjectId);
      if (path == null)
         return false;

      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return false;
      if (!ds.bind())
         return false;

      // open an editing batch
      if (!ds.openBatch(new String(path)))
      {
         ds.unbind();
         return false;
      }

      // set the target directory object's "value" attribute of type Integer
      boolean rc = ds.setNodeByteArray(new String(path + "/" + extName),
                                       "value", value);

      // close the batch
      if (!ds.closeBatch(true))
      {
         ds.unbind();
         return false;
      }

      // deactivate directory services
      ds.unbind();

      return rc;
   }

   /**
    * Checks whether the specified user account exists.
    *
    * @param  uid
    *         user ID
    *
    * @return <code>true</code> if such user account exists
    */
   public boolean isUserDefined(String uid)
   {
      Account acc = getAccount(uid);
      
      // check the account type
      return (acc != null && acc.getAccountType() == Account.ACC_USER);
   }

   /**
    * Checks whether the specified process account exists.
    *
    * @param  pid
    *         process ID
    *
    * @return <code>true</code> if such process account exists
    */
   public boolean isProcessDefined(String pid)
   {
      Account acc = getAccount(pid);

      // check the account type
      return (acc != null && acc.getAccountType() == Account.ACC_PROCESS);
   }

   /**
    * Checks whether the specified group account exists.
    * 
    * @param  uid
    *         group ID
    * 
    * @return <code>true</code> is such group account exists
    */
   public boolean isGroupDefined(String uid)
   {
      Account acc = getAccount(uid);
      
      // check the account type
      return (acc != null && acc.getAccountType() == Account.ACC_GROUP);
   }

   /**
    * Get the accounts configured to start the given appserver.  If the <code>appServer</code> is
    * <code>null</code>, this will return an array of length 0.
    * 
    * @param    appServer
    *           The appserver for which the accounts are needed.
    * 
    * @return   The accounts configured for this appServer, if any.
    */
   public String[] getAccountsForAppserver(String appServer)
   {
      return getCache().getAccountsForAppserver(appServer);
   }

   /**
    * Gets IDs of all defined user accounts.
    *
    * @return array of all defined user IDs
    */
   public String[] getAllUsers()  
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return new String[] {};

      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // get and return tha array
      return sc.getAllAccountIds(Account.ACC_USER);
   }

   /**
    * Gets IDs of all defined accounts.
    *
    * @return array of all defined subject IDs
    */
   public String[] getAllSubjects()  
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return new String[] {};

      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // get and return tha array
      return sc.getAllAccountIds(Account.ACC_ALL);
   }

   /**
    * Gets IDs by thier ordinals.
    *
    * @param ordinal
    *        ordinal number of some subject ID
    * @return a subject ID that corresponds to the ordinal number given.
    */
   public String getIdByOrdinal(int ordinal)  
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return null;

      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // get the account
      Account subj = sc.getAccountByOrd(ordinal);
      if (subj == null)
         return null;
      
      return subj.getSubjectId();
   }

   /**
    * Gets the user ID of a user having a matching value in a specified 
    * named account extension of integer type.
    *
    * @param extName
    *        account extension name
    *
    * @param extValue
    *        extension value 
    *
    * @return user ID or <code>null</code> if no such extension exists.
    */
   public String getIdByExtInteger(String extName, int extValue)
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkExtensionsAccess();
      if (!decision)
         return null;

      // get all user account IDs
      String[] uid = getAllUsers();

      String retValue = null;
      
      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // try each user ID
      String path = null;
      Integer value = null;
      
      for (int i = 0; i < uid.length; i++)
      {
         path = locateExtensionPath(uid[i]);
         if (path == null)
            continue;
   
         // get the directory object's "value" attribute of type Integer
         value = ds.getNodeInteger(new String(path + "/" + extName), "value");
         if (value == null)
         {
            continue;
         }
         
         if (value.intValue() == extValue)
         {
            retValue = uid[i];
            break;
         }
      }
      
      // deactivate directory services
      ds.unbind();

      return retValue;
   }

   /**
    * Gets the user ID of a user having a matching value in a specified 
    * named account extension of string type.
    *
    * @param extName
    *        account extension name
    *
    * @param extValue
    *        extension value 
    *
    * @param cs
    *        case sensitivity for comparison. <code>true</code> means 
    *        case-sensitive comparison. 
    *
    * @return user ID or <code>null</code> if no such extension exists.
    */
   public String getIdByExtString(String extName, String extValue, boolean cs)
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkExtensionsAccess();
      if (!decision)
         return null;

      // get all user account IDs
      String[] uid = getAllUsers();

      String retValue = null;
      
      // activate directory services
      DirectoryService ds = DirectoryService.getInstance();
      if (ds == null)
         return null;
      if (!ds.bind())
         return null;

      // try each user ID
      String path = null;
      String value = null;
      boolean comp = false;
      
      for (int i = 0; i < uid.length; i++)
      {
         path = locateExtensionPath(uid[i]);
         if (path == null)
            continue;
   
         // get the target directory object's "value" attribute of type Integer
         value = ds.getNodeString(new String(path + "/" + extName), "value");
         if (value == null)
         {
            continue;
         }
         
         // compare extension value accounting for requested case sensitivity
         if (cs)
            comp = value.equals(extValue);
         else
            comp = value.compareToIgnoreCase(extValue) == 0;

         if (comp)
         {
            retValue = uid[i];
            break;
         }
      }
      
      // deactivate directory services
      ds.unbind();

      return retValue;
   }

   /**
    * Returns the array of all ACLs defined for the specified resource type.
    * <p>
    *
    * @param  resourceId
    *         resource identity
    * @return array of <code>AccessControlList</code>s. The array is empty if
    *         the caller has no access rights to the system/accounts resource 
    *         instance.
    */
   public AccessControlList[] getACLs(int resourceId) 
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return new AccessControlList[] {};

      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();

      if (resourceId < 0 || resourceId >= nres)
      {
         return new AccessControlList[] {};
      }

      // access the plugins registry and the ACLs
      ResourceRegistry reg = sc.getRegistryByOrd(resourceId);
      AccessControlList[] acls = reg.getACLs();

      // making an array copy to give away
      AccessControlList[] ret = new AccessControlList[acls.length];
      for (int i = 0; i < acls.length; i++)
      {
         ret[i] = acls[i];
      }
      
      return ret;
   }

   /**
    * Get a map of consolidated ACL's, per each function. For each function 
    * name there will be a map of subject indexes to their permissions.
    * 
    * @param   type
    *          The resource type name for which the consolidated ACLs are 
    *          needed
    *          
    * @return  a map of consolidated ACL's, for the specified resource type.
    */
   public Map<String, Map<Integer, Rights>> getConsolidatedACLs(String type)
   {
      AbstractResource resource = getPluginInstance(type);
      
      if (resource == null)
      {
         return Collections.EMPTY_MAP;
      }

      Map<String, Map<Integer, Rights>> result = 
         new HashMap<String, Map<Integer, Rights>>();
      
      // for each instance name, create a set of subjects with explicit 
      // permissions for that instance name
      Map<String, Set<Integer>> subjectsByInstances = 
         getSubjectsByInstanceName(resource.resourceIndex);
      
      SecurityCache sc = getCache();
      
      Iterator<String> iter = subjectsByInstances.keySet().iterator();
      while (iter.hasNext())
      {
         String instanceName = iter.next();
         Set<Integer> subjects = subjectsByInstances.get(instanceName);
         
         Map<Integer, Rights> permissions = result.get(instanceName);
         
         if (permissions == null)
         {
            permissions = new HashMap<Integer, Rights>();
            result.put(instanceName, permissions);
         }
         
         // iterate the known subjects with explicit permissions and compute 
         // the "consolidated" permissions
         for (Integer subjectIdx : subjects)
         {
            Account acc = sc.getAccountByOrd(subjectIdx);
            if (acc == null)
            {
               continue; // not a valid account
            }
            
            Rights rights = resource.getPermissions(acc.getSubjectId(), 
                                                    instanceName);

            if (rights != null)
            {
               permissions.put(subjectIdx, rights);
            }
         }
      }
      
      return result;
   }

   /**
    * Consolidate all the subjects for each instance with defined ACLs, for
    * the given resources. 
    * 
    * @param   resourceId
    *          resource identity
    *          
    * @return  a map which holds a set of subjects for each resource instance
    */
   private Map<String, Set<Integer>> getSubjectsByInstanceName(int resourceId) 
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return Collections.EMPTY_MAP;
      
      // Pick up the generation of the security cache
      SecurityContextStack ctx = SecurityContextStack.getContext();
      SecurityCache sc = ctx.getCache(); 
      int nres = sc.getNumResources();
      
      if (resourceId < 0 || resourceId >= nres)
      {
         return Collections.EMPTY_MAP;
      }
      
      // access the plugins registry and the ACLs
      ResourceRegistry reg = sc.getRegistryByOrd(resourceId);
      AccessControlList[] acls = reg.getACLs();
      
      Map<String, Set<Integer>> result = new HashMap<String, Set<Integer>>();
      
      for (int i = 0; i < acls.length; i++)
      {
         String instanceName = acls[i].getInstanceName();

         Set<Integer> subjects = result.get(instanceName);
         if (subjects == null)
         {
            subjects = new HashSet<Integer>();
            result.put(instanceName, subjects);
         }
         
         // get all bindings for this resource instance
         List binds = acls[i].getBindings();
         
         // collect the subjects
         Iterator iter = binds.iterator();
         while (iter.hasNext())
         {
            Binding bind = (Binding) iter.next();

            int[] boundedSubjects = bind.getSubjects();
            subjects.addAll(Utils.primitiveToIntegerList(boundedSubjects));
         }
      }
      
      return result;
   }

   /**
    * Initiates the password change procedure, calls the configured password
    * input hook, and changes the password for the current user when it is
    * verified.
    *
    * @return  <code>true</code> if password has been changed successfully.
    */
   public boolean changePassword() 
   {
      // Pick up the generation of the security cache
      SecurityCache sc = getCache(); 

      // get the password change hook
      String pwiClassName = sc.getPasswordInput();
      PasswordInput pwi = null;
      
      // create an instance of the class 
      try
      {
         Class pwiClass = Class.forName(pwiClassName);
         pwi = (PasswordInput)pwiClass.newInstance();
      }
      catch (Exception e)
      {
         return false;
      }
         
      if (pwi == null)
         return false;

      // get the current password's digest
      String uid = getUserId();
      Account acc = sc.getAccountById(uid);
      if (acc == null)
         return false;
      if (acc.getAccountType() != Account.ACC_USER)
         return false;
      
      UserAccount user = (UserAccount)acc;

      byte[] hashpw = null;

      synchronized (pwchSync)
      {
         hashpw = user.getPassword();
      }
      
      // run the password input hook
      String newpw = pwi.obtainPassword(hashpw);
      if (newpw == null)
         return false;

      // set new password
      return setPassword(newpw);
   }

   /**
    * Gets password age status for the current account.
    *
    * @return <code>true</code> if this password is too old and needs changing
    */
   public boolean isPasswordAged()
   {
      // Pick up the generation of the security cache
      SecurityCache sc = getCache(); 

      // get the current account
      String uid = getUserId();
      Account acc = sc.getAccountById(uid);
      if (acc == null)
         return false;
      if (acc.getAccountType() != Account.ACC_USER)
         return false;
      
      UserAccount user = (UserAccount)acc;
      return user.isPasswordAged();
   }

   /**
    * Returns the string representation of this object.
    *
    * @return string describing this object
    */
   public String toString()
   {
      return "Security Manager";
   }

   /**
    * Adds a token to the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @param token
    *        an arbitrary object to be kept in the entry
    *
    * @return <code>true</code> if the token is added successfully
    */
   boolean addToken(Object key, Object token)
   {
      SecurityContextStack ctx = SecurityContextStack.getContext();
      return (ctx != null && ctx.addToken(key, token));
   }

   /**
    * Remove a token from the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return <code>true</code> if the token was removed successfully
    */
   boolean removeToken(Object key)
   {
      SecurityContextStack ctx = SecurityContextStack.getContext();
      return (ctx != null && ctx.removeToken(key));
   }

   /**
    * Checks whether the specified token is in the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return <code>true</code> if the token is found
    */
   public boolean hasToken(Object key)
   {
      SecurityContextStack ctx = SecurityContextStack.getContext();
      return (ctx != null && ctx.hasToken(key));
   }

   /**
    * Gets the token from the context map.
    *
    * @param key
    *        token key to be used as a key in the map
    *
    * @return the value of the token as a generic object or <code>null</code>
    */
   public Object getToken(Object key)
   {
      SecurityContextStack ctx = SecurityContextStack.getContext();
      return (ctx == null ? null : ctx.getToken(key));
   }

   /**
    * Safely gets the reference to the security cache in a multithreaded
    * environment.
    * @return reference to the curent copy of <code>SecurityCache</code>
    */
   synchronized SecurityCache getCache()  
   {
      return cache;
   }

   /**                              
    * Forcibly terminates (closes) the socket in use for the given session. At
    * this time it is not possible to terminate virtual sessions.
    *
    * @param    sid
    *           Session ID to terminate.
    *
    * @throws   IllegalStateException
    *           If the given session id is invalid, is associated with a
    *           virtual session or if it is associated with the current
    *           session being used.
    */
   void killSession(int sid)
   {
      Iterator  iter    = null;
      Session   session = null;
      NetSocket socket  = null;
      
      SecurityContextStack stack   = SecurityContextStack.getContext();
      SecurityContext      current = stack.getEffectiveContext();
      int                  thisId  = current.getSessionId();
      
      // make sure we don't kill the current/active session
      if (sid == thisId)
      {
         throw new IllegalStateException("You cannot terminate the session " +
                                         "which you are currently using!");
      }
      
      synchronized (listOfSess)
      {
         iter = listOfSess.listIterator();
         
         while (iter.hasNext())
         {
            session = (Session) iter.next();

            SecurityContext context = session.getSecurityContext();
            
            if (context.getSessionId() == sid)
            {
               Object id = session.getSessionId();
               
               if (id instanceof NetSocket)
               {
                  socket = (NetSocket) id;
                  break;
               }
               else
               {
                  String spec = "Session ID %d cannot be killed. It may be a" +
                                " virtual session (not killable at this time).";
                  throw new IllegalStateException(String.format(spec, sid));
               }
            }
         }
      }
      
      if (socket == null)
      {
         throw new IllegalStateException(String.format("Invalid session ID %d!",
                                                       sid));
      }
      
      // kill the socket, this will cause both sides to "unwind" and disconnect
      socket.close();
   }

   /**
    * Get the list of all active sessions which have executed API(s)
    * associated to this jar.
    *
    * @param   jar
    *          The name of the target jar file.
    * @param   iface
    *          The interface for which we need the sessions; 
    *          if <code>null</code>, search through the entire jar.
    *
    * @return  the list of all active sessions which have executed API(s)
    *          associated to this jar.
    */
   SessionInfo[] getActiveSessions(String jar, String iface)
   {
      Object[] contexts = SessionManager.get().getSessionContexts(jar, iface);
      List<SessionInfo> data = new ArrayList<SessionInfo>(contexts.length);

      for (Object o : contexts)
      {
         Session sess = locateSession((SecurityContext) o);
         SessionInfo descr = getSessionDescriptor(sess);
         
         data.add(descr);
      }
      
      return data.toArray(new SessionInfo[0]);
   }

   /**
    * Terminate all active sessions which have invoked the given APIs in the
    * given jar. If the jar name is <code>null</code>, terminate all active
    * tracked sessions.
    * This method guarantees that all session termination listeners have been
    * executed.
    *
    * @param jar   The jar name; if <code>null</code>, terminate all active
    *              tracked sessions.
    * @param iface The interface for which we need to terminate the sessions;
    *              if <code>null</code>, terminate all sessions for the given
    *              jar.
    */
   void terminateSessions(String jar, String iface)
   {
      try
      {
         SessionManager.get().terminateSessions(jar, iface);
      }
      catch (RestrictedUseException e)
      {
         throw new IllegalStateException("Call of " +
               "SessionManager.terminateSessions from this function should" +
               "be granted.");
      }
   }

   /**
    * Creates a report of all current user and process sessions.
    *
    * @return   Session descriptors suitable for reporting purposes.
    */
   SessionInfo[] getSessionReport()  
   {
      Iterator          iter = null;
      Session           sess = null;
      List<SessionInfo> info = new ArrayList<SessionInfo>();
      
      synchronized (listOfSess)
      {
         iter = listOfSess.listIterator();
         
         while (iter.hasNext())
         {
            sess = (Session) iter.next();
            
            SessionInfo descr = getSessionDescriptor(sess);
            
            // virtual session cannot be inspected at this time, so avoid
            // listing virtual sessions
            if (descr != null)
               info.add(descr);
         }
      }

      return info.toArray(new SessionInfo[0]);
   }
   
   /**
    * Locates the session in the list of sessions with matching security 
    * context.
    *
    * @param  context
    *         security context to look for
    * @return instance of <code>Session</code> associated with the security
    *         context
    */
   Session locateSession(SecurityContext context)  
   {
      Iterator iter = null;
      Session  sess = null;
      
      // purge the session data
      synchronized (listOfSess)
      {
         iter = listOfSess.listIterator();
         
         while (iter.hasNext())
         {
            sess = (Session) iter.next();
            
            if (sess.hasContext(context))
               break;
         }
      }

      return sess;
   }  

   /**
    * Obtain a descriptor for the given session.  At this time virtual sessions
    * cannot be inspected.
    *
    * @param    session
    *           The session on which to report.  Must not be <code>null</code>.
    *
    * @return   The descriptor, suitable for reporting on the session details.
    *           <code>null</code> will be returned if the session cannot be
    *           inspected.
    *
    * @throws   NullPointerException
    *           If the given session is <code>null</code>.
    */
   SessionInfo getSessionDescriptor(Session session)
   {
      if (session == null)
         throw new NullPointerException("Session may not be null!");
      
      Object    id     = session.getSessionId();
      NetSocket socket = null;
      
      if (id instanceof NetSocket)
      {
         socket = (NetSocket) id;
      }
      else
      {
         return null;
      }
      
      SessionInfo info = new SessionInfo();
      
      SecurityContext context = session.getSecurityContext();
      SecurityCache   cache   = session.getCache();
      
      // read the user ordinal
      int ord = context.getUserOrdinal();
      
      if (ord == -1)
      {
         // we aren't a user, perhaps we are a process?
         info.user = false;
         ord = context.getProcessOrdinal();
      }
      
      Account acct = cache.getAccountByOrd(ord); 
      
      info.name      = acct.getSubjectId();
      info.descr     = acct.getDescription();
      info.id        = context.getSessionId();
      info.remoteNet = socket.getRemoteSockAddr().toString();
      info.localNet  = socket.getLocalSockAddr().toString();
      info.secure    = (socket.getSession() != null);
      info.logonTime = session.getTimeStamp();
      info.cacheGen  = cache.getSerial();
      
      return info;
   }

   /**
    * Returns the bootstrap configuration used to construct this instance.
    *
    * @return instance of <code>BootstrapConfig</code> associated with this
    *         object
    */
   BootstrapConfig getConfig()
   {
      return cfg;
   }

   /**
    * Checks the rights of the caller to perform admin operations.
    *
    * @return <code>true</code> if the caller has sufficient rights
    */
   private boolean adminAccess()
   {
      // check the access rights with the admin plugin first
      AdminResource ar = (AdminResource)getPluginInstance("admin");
      if (ar != null)
      {
         return ar.checkAnyAdminAccess();
      }
      
      // fall back to the system resource
      SystemResource sr = (SystemResource)getPluginInstance("system");
      return sr.checkAdminAccess();
   }

   /**
    * Get the account associated with the given user id.
    * 
    * @param  uid
    *         user ID
    * 
    * @return The account, if it exists, and the caller has sufficient rights
    *         to retrieve it, else <code>null</code>
    */
   private Account getAccount(String uid)
   {
      // check the access rights
      SystemResource sr = (SystemResource)getPluginInstance("system");
      boolean decision = sr.checkAccountsAccess();
      if (!decision)
         return null;

      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // find the account
      Account acc = sc.getAccountById(uid);
      
      return acc;
   }

   /**
    * Creates an audit record for security context creation event. 
    *
    * @param  cache
    *         security cache where to log this event
    *
    * @param  ctx
    *         security context
    */
   private void logNewContext(SecurityCache cache, SecurityContext ctx)
   {
      Audit audit = cache.getAudit();
      if (!audit.isEnabled())
      {
         return;
      }
      
      StringBuffer sb = new StringBuffer();
      int[] checkList = ctx.getCheckList();

      for (int i = 0; i < checkList.length; i++)
      {
         sb.append(checkList[i]);
         sb.append("=");
         sb.append(cache.getAccountByOrd(checkList[i]).getSubjectId());
         sb.append(":");
      }
      sb.setLength(sb.length() - 1);

      audit.log(cache.getSystem(), "context", 
                SystemResource.SYSTEM_CONTEXT_CREATE, true, new String(sb));
   }

   /**
    * Creates an audit record for security context switch event. 
    *
    * @param  cache
    *         security cache where to log this event
    *
    * @param  ctx
    *         coming security context, may be <code>null</code>
    */
   private void logContext(SecurityCache cache, SecurityContext ctx)
   {
      Audit audit = cache.getAudit();
      
      if (!audit.isEnabled())
      {
         return;
      }
      
      StringBuffer sb = new StringBuffer();
      int[] checkList = ctx == null ? new int[0] : ctx.getCheckList();

      for (int i = 0; i < checkList.length; i++)
      {
         sb.append(checkList[i]);
         sb.append("=");
         sb.append(cache.getAccountByOrd(checkList[i]).getSubjectId());
         sb.append(":");
      }
      if (sb.length() > 0)
         sb.setLength(sb.length() - 1);

      audit.log(cache.getSystem(),
                "context", 
                SystemResource.SYSTEM_CONTEXT_SWITCH,
                true,
                sb.toString());
   }

   /**
    * Creates an audit record for a terminated security context. 
    *
    * @param  ctx
    *         security context
    */
   private void logGoingContext(SecurityContext ctx)
   {
      SecurityCache sc = getCache();
      Audit audit = sc.getAudit();
      if (!audit.isEnabled())
      {
         return;
      }
      
      StringBuffer sb = new StringBuffer();
      int[] checkList = ctx.getCheckList();

      for (int i = 0; i < checkList.length; i++)
      {
         sb.append(checkList[i]);
         sb.append("=");
         sb.append(getCache().getAccountByOrd(checkList[i]).getSubjectId());
         sb.append(":");
      }
      sb.setLength(sb.length() - 1);

      audit.log(sc.getSystem(), "context", 
                SystemResource.SYSTEM_CONTEXT_DELETE, true, new String(sb));
   }

   /**
    * Creates an audit record for a resource instance access event. 
    *
    * @param  cache
    *         security cache where to log this event
    *
    * @param  resourceId
    *         resource identity
    *
    * @param  instanceName
    *         resource instance name
    *
    * @param  mode
    *         access mode (requested rights) the decision was made about
    *
    * @param  decision
    *         access decision
    *
    * @param  message
    *         additional text 
    */
   void logResource(SecurityCache cache, int resourceId, String instanceName,  
                    int mode, boolean decision, String message)
   {
      Audit audit = cache.getAudit();
      audit.log(resourceId, instanceName, mode, decision, message); 
   }

   /**
    * Creates an audit record for a resource instance access event. 
    *
    * @param  cache
    *         security cache where to log this event
    *
    * @param  resourceId
    *         resource identity
    *
    * @param  instanceName
    *         resource instance name
    *
    * @param  mode
    *         access mode (requested rights) the decision was made about
    *
    * @param  decision
    *         access decision
    */
   void logResource(SecurityCache cache, int resourceId, String instanceName,  
                    int mode, boolean decision)
   {
      Audit audit = cache.getAudit();
      audit.log(resourceId, instanceName, mode, decision); 
   }

   /**
    * Gets the requested debug level.
    * 
    * @return requested debug level 
    */
   Level getRequestedDebugLevel() 
   {
      return debugLevelRequest; 
   }
   
   /**
    * Sets new password for the current user account.
    *
    * @param   newPassword
    *          new password as plain text
    * @return  <code>true</code> if password has been set successfully
    */
   boolean setPassword(String newPassword) 
   {
      // find the security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext      ctx  = ctxs.getEffectiveContext();

      // figure out 1st subject ordinal
      int ord = ctx.getUserOrdinal();
      
      if (ord == -1)
      {
         // we aren't a user! sorry, no password change for processes
         return false;
      }

      PasswordChange pwch = new PasswordChange(ord, newPassword);
      Thread thread;
      thread = new Thread(secThreads, pwch, "passwordchange");

      synchronized (pwchSync)
      {
         thread.start();
         try
         {
            thread.join();
         }
         catch (InterruptedException ix)
         {
            return false;
         }
      }
      
      return pwch.result;
   }
   
   /**
    * Validate the server's certificate if this was specified in the 
    * bootstrap configuration.
    *
    * @param   locSess
    *          The local SSL session to validate.
    * @param   config
    *          BootstrapConfig instance which should override the security
    *          manager's version;  <code>null</code> to use default config.
    */
   private boolean validateServer(SSLSession locSess, BootstrapConfig config)
   {
      java.security.cert.Certificate[] peerCerts = null;
      
      // TODO: is getPeerCertificates() required if needsServerValidation()
      //       returns false or can it be deferred?
      
      // query server's certificate
      try
      {
         peerCerts = locSess.getPeerCertificates();
      }
      catch (Exception e)
      {
         // exception is handled as if there was 0 certificates read
      }
      
      // the following server certificate validation steps
      // are taken only if implicitly or explicitly requested

      if (needsServerValidation(config))
      {
         // fail authentication if no server certificate is available 
         if (peerCerts == null || peerCerts.length == 0)
         {
            return false;
         }
         
         X509Certificate servCert = (X509Certificate) peerCerts[0];
   
         // validate the certificate
         try
         {
            servCert.checkValidity();
         }
         
         catch (Exception e)
         {
            return false;
         }
         
         // TODO: what is the difference between the checkValidity() above
         //       and this name based checking below, why are both done?
         //       (document this as it is not clear at all)
         
         List names = null;
         
         // build a set of DNS names/common names to check
         try
         {
            names = fetchAltNames(servCert, 2);
         }
         
         catch (CertificateParsingException e)
         {
            // this exception can't be tolerated
            return false;
         }
   
         // get CN if no altNames
         if (names.isEmpty())
         {
            try
            {
               names.add(getSubjectCommonName(servCert));
            }
            
            catch (CertificateParsingException e)
            {
               // this exception can't be tolerated
               return false;
            }
         }
   
         // compare known names with the used hostname
         int i = hostMatch(locSess.getPeerHost(), names);
         
         if (i == -1)
         {
            // all other types of checking failed, if a truststore alias
            // was specified, then lookup the certificate by that alias and
            // compare it with the server's certificate
            
            TransportSecurity ts = null;
            try
            {
               // if this is the default config, use the default transport
               // security
               ts = (config == this.cfg) ? tranSec 
                                         : getTransportSecurity(config);
            }
            catch (Exception e) 
            {
               return false;
            }
            
            KeyStore trust = ts.getTrustStore();
            
            try
            {
               String trustAlias = config.getConfigItem("security",
                                                        "truststore",
                                                        "alias");
               X509Certificate trusted = (X509Certificate)
                                         trust.getCertificate(trustAlias);
                                         
               if (!servCert.equals(trusted))
               {
                  return false;
               }
            }
            
            catch (Exception e)
            {
               return false;
            }
         }
      }
      
      return true;
   }
   
   /**
    * Temporarily switches to a user context to serve the associated request.  
    * The calling thread has to have the initial security context.
    * <p>
    * The use count for the target security context is incremented by 1.
    *
    * @param    key
    *           security context key
    *
    * @throws   RestrictedUseException
    *           If called improperly.
    */
   void pushContextWorker(Object key)
   throws RestrictedUseException
   {
      // check the passed object
      if (!(key instanceof SecurityContext))
         throw new RestrictedUseException("invalid security context key");

      // check the current context
      SecurityContext ctx = (SecurityContext)key;
      SecurityContextStack cts = SecurityContextStack.getContext();
      if (cts == null)
         throw new RestrictedUseException("no initial context established");

      if (!cts.push(ctx))
         throw new RestrictedUseException("illegal context switch");

      // safely query cached data and log the context switch event
      SecurityCache sc = getCache(); 
      logContext(sc, ctx);

      assignContext(ctx);
   }

   /**
    * Restores the initial security context of the calling thread when it is
    * done with the user request.  The calling thread has to have its security
    * context switched to a user context prior to this call.
    * <p>
    * The use count for the current security context is decremented by 1.  
    * If it becomes 0, it gets deleted immediately.
    *
    * @throws   RestrictedUseException
    *           If called improperly.
    */
   private void popContextWorker()
   throws RestrictedUseException
   {
      SecurityContextStack cts = SecurityContextStack.getContext();
      
      if (cts == null)
         throw new RestrictedUseException("no security context established");

      SecurityContext ctx = cts.pop();
      
      if (ctx == null)
         throw new RestrictedUseException("illegal context switch");
      
      endContext(cts, ctx);
   }
   
   /**
    * Worker routine to centralize unassignment and cleanup (if this was the
    * final usage) of the given context.
    *
    * @param    cts
    *           The context stack for the current thread.
    * @param    ctx
    *           The context to end.
    */
   private void endContext(SecurityContextStack cts, SecurityContext ctx)
   {
      // unassign context and do cleanup if it's time
      boolean finalUse = unassignContext(ctx);
      
      if (finalUse)
      {
         // IMPORTANT NOTE:
         // The simpler approach seems to be synchronizing on the 
         // SecurityContext instance itself (and making any other conflicting
         // threads wait until cleanup is done), but this must not be done,
         // as it will deadlock.
         // Consider this scenario: Conversation thread starts the context 
         // cleanup code (acquiring a lock on the SecurityContext instance). 
         // The cleanup code for ConnectionManager will send the command to 
         // end the virtual session and will wait for this command to 
         // complete. When the Reader thread for the virtual session notices 
         // that it needs to end, it will stop the queue and will AGAIN start
         // the context cleanup code for the same SecurityContext instance 
         // (as the Reader for the virtual session is closed using the context
         // of the initiator). As the Conversation thread has a lock on this,
         // Reader can not go forward until Conversation thread releases the 
         // lock. But Conversation thread is still in waiting mode to receive
         // an answer for its "end virtual session" command... so, deadlock.

         boolean clean = true;
         synchronized (cleanupLocks)
         {
            clean = (cleanupLocks.get(ctx) == null);
            if (clean)
            {
               cleanupLocks.put(ctx, true);
            }
         }

         if (clean)
         {
            // context cleanup code is invoked only if we are the first thread
            // which executes this code; else, skip it, as another thread is
            // doing this work
            try
            {
               // non-initial contexts have already been popped by this time,
               // push it back to ensure the context is valid for the lifespan
               // of cleanup
               if (!ctx.isInitial())
               {
                  cts.push(ctx);
               }
               
               ctx.cleanup();
      
               if (!ctx.isInitial())
               {
                  cts.pop();
               }
      
               logGoingContext(ctx);
            }

            finally
            {
               synchronized (cleanupLocks)
               {
                  cleanupLocks.remove(ctx);
               }
            }
         }
      }
      
      // safely query cached data and log the context switch event
      SecurityCache sc = getCache();
      logContext(sc, ctx);
   }

   /**
    * Creates an instance of <code>TransportSecurity</code> class that serves
    * the client or server end of secured TLS connections, as needed.  The
    * newly created instance is a singleton within the domain of this 
    * instance of <code>SecurityManager</code>. Thus, repeatable calls return
    * the same reference.
    *
    * @return   The instance that configures the SSL environment with custom 
    *           key stores and trust stores.
    */
   private synchronized TransportSecurity getTransportSecurity(BootstrapConfig config)
   throws NoSuchAlgorithmException,
          KeyStoreException, 
          UnrecoverableKeyException,
          ConfigurationException
   {
      TransportSecurity ts = tranSec;
      
      // if the default transport security was not initialized or a different
      // bootstrap config is used (i.e. not the default config), then create
      // the transport security instance
      if (tranSec == null || config != this.cfg)
      {
         boolean srv = config.isServer();
         
         KeyStore trust = null;
         
         if (srv)
         {
            // safely query cached data
            SecurityCache sc = getCache(); 
   
            // obtain the cached trust store
            trust = sc.getTrustStore();
         }
         
         ts = new TransportSecurity(config, srv, trust);
         
         if (config == this.cfg)
         {
            // set the default transport security, if this is the default
            // bootstrap config
            tranSec = ts;
         }
      }
      
      return ts;
   }
   
   /**
    * Detects if server certificate validation is required.
    *
    * @param   config
    *          BootstrapConfig instance which should override the security
    *          manager's version;  <code>null</code> to use default config.
    *          
    * @return  <code>true</code> if server certificate validation is
    *          required.
    */
   private boolean needsServerValidation(BootstrapConfig config)
   {
      if (config == null)
      {
         config = this.cfg;
      }
      
      String item = null;
      
      try
      {
         item = config.getConfigItem("security", "certificate", "validate");
      }
      
      catch (ConfigurationException cfg)
      {
         // assume no server cert validation
      }

      return "true".equals(item);
   }
      
   /**
    * Send the client's requested authorization type on the given connection.
    * <p>
    * This method inspects the bootstrap configuration to see what kind of
    * authentication to pick:<ul>
    *   <li> <code>security.keystore.processalias</code> is set; 
    *        <code>AUTH_REQ_PROCESS</code> will be sent.
    *   <li> <code>security.keystore.processalias</code> is not set, but
    *        <code>security.authentication.type</code> is set and equals
    *        <code>PROGRAM</code>;
    *        <code>AUTH_REQ_PROGRAM</code> will be sent.
    *   <li> otherwise 
    *        <code>AUTH_REQ_USER</code> will be sent.
    *
    * @param    output
    *           The connection on which to write our type.
    * @param    cfg
    *           The attempted connection's configuration.
    *
    * @return   <code>true</code> if successful
    */
   private boolean sendAuthType(ObjectOutputStream output, BootstrapConfig cfg)
   {
      // authentication type to be determined dynamically
      int authType = -1;
      
      try
      {
         String alias = cfg.getConfigItem("security",
                                          "keystore",
                                          "processalias");
         
         if (alias != null)
         {
            // use process type authentication
            authType = AUTH_REQ_PROCESS;
         }
      }
      
      catch (ConfigurationException ce)
      {
         // authentication type not determined yet
         authType = -1;
      }

      // choose between interactive and programmatic types if not process
      if (authType == -1)
      {
         // check to see if security.authentication.type is set
         try
         {
            String type = cfg.getConfigItem("security",
                                            "authentication",
                                            "type");
            
            if (type != null && type.equalsIgnoreCase("program"))
            {
               // use programmatic type authentication
               authType = AUTH_REQ_PROGRAM;
            }
         }
         
         catch (ConfigurationException ce)
         {
            // authentication type not determined yet
            authType = -1;
         }
      }

      if (authType == -1)
      {
         authType = AUTH_REQ_USER;
      }

      dTrace("Sending AUTHTYPE: " + authType);
      try
      {
         // send the packet
         output.writeInt(authType);
         if (authType == AUTH_REQ_USER)
         {
            String userName = "";
            try
            {
               userName = cfg.getConfigItem("access", "subject", "id");
               if (userName == null)
                  userName = "";
            }
            catch (ConfigurationException e)
            {
               dWarn("Error accessing subject id: " + e);
            }

            output.writeUTF(userName);
         }

         output.flush();
      }
      
      catch (IOException ioe)
      {
         return false;
      }
      
      dTrace("Sent AUTHTYPE");
      
      return true;
   }
   
   /**
    * Verifies the plain text password against the one stored with 
    * the specified user account.
    *
    * @param   plain
    *          plain text password
    * @param   user
    *          <code>UserAccount</code> instance
    * @return  <code>true</code> if password is valid
    */
   private boolean checkPassword(String plain, UserAccount user)
   {
      if (!user.isProtected())
      {
         // this user account is not protected with a password
         return true;
      }
      
      byte[] hashpw = null;

      synchronized (pwchSync)
      {
         hashpw = user.getPassword();
      }

      if (hashpw == null)
      {
         // no password assigned - the account can't be used
         return false;
      }

      byte[] testpw = HashPassword.hashPassword(plain);
      
      if (testpw == null)
      {
         return false;
      }

      if (testpw.length != hashpw.length)
      {
         dTrace("Wrong password");
         return false;
      }
      
      for (int i = 0; i < hashpw.length; i++)
      {
         if (testpw[i] != hashpw[i])
            {
               dTrace("Wrong password");
               return false;
            }
      }
      
      return true;
   }
   
   /**
    * Calculates the age of the existing user account password and checks to
    * see whether it is too old, according to the maxAge parameter. 
    *
    * @param   user
    *          <code>UserAccount</code> instance
    */
   private void agePassword(UserAccount user)
   {
      // Pick up the generation of the security cache
      SecurityCache sc = getCache();

      // get and check the password aging parameter
      int max = sc.getMaxAge();
      if (max == 0)
         return;

      // get the timestamp of the last password change, if any
      DateValue pdate = user.getPasswordDate();
      TimeValue ptime = user.getPasswordTime();

      // no or incomplete timestamp means the password is considered aged
      if (pdate == null || ptime == null)
      {
         user.setPasswordAged(true);
         return;
      }
         
      // calculate password age
      Date pwDate = pdate.getDate();
      Date now = new Date();
      
      long maxAge = ((long)max) * 86400 * 1000;
      long age = now.getTime() - pwDate.getTime();
      
      // flag the password as aged if its age exceeds the limit
      if (age > maxAge)
         user.setPasswordAged(true);
   }

   /**
    * Assign the current thread to the given security context, and add a
    * unique identifier for this thread to the thread ID map.
    *
    * @param   context
    *          Security context to be assigned.
    */
   private void assignContext(SecurityContext context)
   {
      context.assign();

      synchronized (threadIDs)
      {
         if (!threadIDs.containsKey(Thread.currentThread()))
         {
            threadIDs.put(Thread.currentThread(), 
                          new Integer(++nextThreadID));
         }
      }
   }
   
   /**
    * Unassign the current thread from the given security context, and remove
    * its entry from the thread ID map.
    *
    * @param   context
    *          Security context to be unassigned.
    */
   private boolean unassignContext(SecurityContext context)
   {
      return context.unassign();
   }
   
   /**
    * Safely sets the reference to the security cache in a mutithreaded
    * environment.
    * <p>
    * In the multithreaded environment, after the closing of a batch editing
    * session, a newly created thread rereads the directory and builds a new
    * instance of <code>SecurityCache</code> while the existing one continues
    * to serve the needs of the server. This call is the last step in the
    * refreshing of the security cache.
    *
    * @param  cache
    *         reference to a copy of <code>SecurityCache<code>
    */
   private synchronized void setCache(SecurityCache cache)  
   {
      this.cache = cache;
   }
   
   /**
    * Instantiates and runs the server side of the custom authentication hook.
    * <p>
    * Accepts the byte array produced by the client side authorization hook
    * as the authentication input, and custom parameters.
    *
    * @param    sc
    *           The security cache generation to be used.
    * @param    targetHookName
    *           The name of the custom auth hook to run as it is represented
    *           into the directory. Cannot be <code>null</code>.  
    * @param    serverHookName
    *           The class name of the server hook or <code>null</code> to
    *           obtain the default processing.
    * @param    parm
    *           Custom parameters.
    * @param    auth
    *           The authorization input from the client.
    *
    * @return   The result of the authentication processing.
    */
   private AuthenticationResponse runCustomServerHook(SecurityCache sc,
                                                      String        targetHookName,
                                                      String        serverHookName,
                                                      String        parm,
                                                      byte[]        auth)
   {
      AuthenticationResponse result = null;

      // create an instance of the class and call serverAuthHook 
      try
      {
         Authenticator instance = null;
         
         if (serverHookName != null)
         {
            try
            {
               Class server = Class.forName(serverHookName);
               instance = (Authenticator) server.newInstance();
            }
            
            catch (Exception exc)
            {
               // can't find or can't use the user-specified replacement, so
               // use the default approach
               instance = this;
            }
         }
         else
         {
            // no user-specified replacement, so use the default approach
            instance = this;
         }
         
         result = instance.serverAuthHook(auth, parm);
         
         if (result.getReturnCode() != AUTH_RESULT_SUCCESS)    
            return result;
      }
      
      catch (Exception e)
      {
         // something unexpected happened
         return new AuthenticationResponse(null,
                                           AUTH_RESULT_UNSPECIFIED_FAILURE);
      }

      String userId = result.getSubject();
      dTrace("Received user ID <" + userId + ">");
      
      // locate the account
      Account acc = sc.getAccountById(userId);
      
      if (acc == null)
      {
         dTrace("No account for ID <" + userId + ">");
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }
      
      // TODO: Why is this restriction needed? By doing this, one cannot
      //       use the custom auth method to authenticate processes, right?
      // (before changing this check TrustedClientPlugin behavior)
      if (acc.getAccountType() != Account.ACC_USER)
      {
         dTrace("Wrong account type for ID <" + userId + ">");
         return new AuthenticationResponse(null,
                                           AUTH_RESULT_UNSPECIFIED_FAILURE);
      }

      UserAccount userAccount = (UserAccount) acc;
      if (userAccount.getAuthMode() != AUTH_MODE_CUSTOM)
      {
         dTrace("Account <" + userId + "> uses " + userAccount.getAuthMode() +
                " auth mode, cancelling custom auth");
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }

      if (userAccount.getAuthPlugin() == null ||
          !userAccount.getAuthPlugin().equals(targetHookName))
      {
         dTrace("Account <" + userId + "> uses " +
                userAccount.getAuthPlugin() +
                " auth plugin, cancelling custom auth");
         return new AuthenticationResponse(null, AUTH_RESULT_INVALID_USERID);
      }

      return result;
   }  

   /**
    * Private worker method that creates a new security context.
    *
    * @param  sc
    *         <code>SecurityCache</code> where new security context belongs
    * @param  pid
    *         process ID for new security context
    * @param  uid
    *         user ID for new security context
    * @return newly created <code>SecurityContext</code>
    */
   private SecurityContext createSecurityContext(SecurityCache sc, String pid, String uid)  
   {
      int procI = sc.getOrdinalById(pid);
      int userI = sc.getOrdinalById(uid);
      if (procI == -1 && userI == -1)
         return null;
      
      int[] grpsI = null;
      UserAccount uacc = (UserAccount)sc.getAccountByOrd(userI);
      if (uacc != null)
         grpsI = uacc.getGroups();
      int nres  = sc.getNumResources();
      
      int sessionID  = getNextUserSessionID();
      String subject = userI != -1 ? uid : pid;

      SecurityContext ctx = new SecurityContext(sc,
                                                procI,
                                                userI,
                                                grpsI, 
                                                nres,
                                                sessionID,
                                                subject);
      logNewContext(sc, ctx);
      
      return ctx;
   }
   
   /**
    * Retrieve a unique user session ID to assign to a new security context.
    *
    * @return  Unique session ID which is not the reserved, system ID.
    */
   private synchronized int getNextUserSessionID()
   {
      while (++nextUserSessionID == SecurityContext.SYSTEM_SESSION);
      
      return nextUserSessionID;
   }

   /**
    * Locates the account extension directory path for the specified user ID.
    *
    * @param  uid
    *         user ID
    *
    * @return directory path to the account extensions
    */
   private String locateExtensionPath(String uid)  
   {
      // Pick up the generation of security cache
      SecurityContextStack ctxs = SecurityContextStack.getContext();
      SecurityContext ctx = ctxs.getEffectiveContext();
      SecurityCache sc = ctx.getCache(); 

      // find the account
      Account acc = sc.getAccountById(uid);
      if (acc == null)
         return null;

      // figure out account type and extension path
      switch (acc.getAccountType())
      {
         case Account.ACC_USER:
            return new String("/security/accounts/users/" + uid);
         case Account.ACC_GROUP:
            return null;
         case Account.ACC_PROCESS:
            return new String("/security/accounts/processes/" + uid);
      }

      return null;
   }
   
   /**
    * Worker to send back the authentication result and a disposition code.
    *
    * @param    out
    *           The socket to send on.
    * @param    authResult
    *           The authentication result.
    * @param    authAction
    *           The authentication disposition code.
    *
    * @return   <code>true</code> if sent successfully.
    */
   private boolean sendAuthResult(ObjectOutputStream out,
                                  int                authResult,
                                  int                authAction)
   {
      if (isTraceLevel())
      {
         dTrace("Sending AUTHRESULT" + authResult + " (action = " +
                authAction + ")");
      }
             
      try
      {
         // send the packet
         out.writeInt(authResult);
         out.writeInt(authAction);
         out.flush();
      }
      
      catch (IOException ioe)
      {
         if (isErrLevel())
         {
            dErr("Error sending AUTHRESULT " + ioe);
         }
         return false;
      }
      
      if (isTraceLevel())
      {
         dTrace("Sent AUTHRESULT");
      }
      
      return true;
   }

   /**
    * Get the default "option" value (specified into directory) for the given
    * custom auth hook.
    *
    * @param   cache
    *          The security cache generation to be used.
    * @param   hookName
    *          The name of the target custom hook.
    *
    * @return  the "option" value for the given custom auth hook.
    */
   private String getHookParam(SecurityCache cache, String hookName)
   {
      AuthPlugin plugin = cache.getAuthPlugin(hookName);
      return plugin != null ? plugin.getOption() : null; 
   }

   /**
    * Implements concurrent security cache refresh function.
    */
   private class SecurityCacheRefresh
   implements Runnable 
   {
      /** logging mode for the new instance of cache */
      private int lm = 0;
      
      /** buffer to receive cache creation messages */
      
      private List<String[]> msgBuf = null;
      
      /**
       * Serves as the entry point for a separate thread.
       *
       * @param lm
       *        logging mode: one of {LM_ADMIN, LM_CONSOLE, LM_DUAL}
       */
      private SecurityCacheRefresh(int lm, List<String[]> msgBuf)
      {
         this.lm = lm;
         this.msgBuf = msgBuf;
      }
      
      /**
       * Serves as the entry point for a separate thread.
       */
      public void run()
      {
         // Safely query cached data
         SecurityCache sc = getCache(); 
   
         // establish initial security context
         SecurityContext context = sc.getInitialSecurityContext();
         logContext(sc, context);
         SecurityContextStack.setContext(new SecurityContextStack(context));
         assignContext(context);

         DirectoryService ds = DirectoryService.getInstance();
         ds.bind();
         
         // instantiate the security cache
         SecurityCache newCache = null;
         
         synchronized (SecurityManager.this)
         {
            try
            {
               newCache = new SecurityCache(cfg, cacheSerial + 1, lm, msgBuf);
               cacheSerial++;
               dWarn("Security Cache generation " + cacheSerial + 
                      " created.");
            }
            catch (Exception exc)
            {
               dErr("Security Cache generation " + (cacheSerial + 1) + 
                      " failed; " + exc);
               exc.printStackTrace();
            }
            
            if (newCache != null)
            {
               SecurityCache oldCache = getCache(); 
               Audit audit = oldCache.getAudit();
   
               // synchronization below is required between this block and
               // start(), stop() and log() methods in Audit, so that there is 
               // no interruption in audit logging
   
               synchronized (audit)
               {
                  audit.stop();
                  setCache(newCache);
                  audit = newCache.getAudit();
                  try
                  {
                     audit.start();
                  }
                  catch (IOException exc)
                  {
                     // The newly defined audit log file can't be activated and
                     // the old audit log's gone.
                     dErr("*** Audit log failure *** " + exc);
                  }
               }
   
               SecurityContext ctx = newCache.getInitialSecurityContext();
               logNewContext(newCache, ctx);
   
               newCache = null;
            }
         }

         ds.unbind();
         unassignContext(context);
      }
   }

   /**
    * Implements password change that requires the privilege of running within
    * the initial security context.
    */
   private class PasswordChange
   implements Runnable 
   {
      /* user account ordinal to change password for */
      private int ord = -1;
      
      /* new password as plain text */
      private String newPassword = null;

      /* the result of the password change */
      boolean result = false;
      
      /**
       * Constructor.
       *
       * @param ord
       *        user account ordinal to change password for 
       * @param newPassword
       *        new password as plain text
       */
      PasswordChange(int ord, String newPassword)
      {
         this.ord = ord;
         this.newPassword = newPassword;
         this.result = false;
      }

      /**
       * Serves as the entry point for a separate thread.
       */
      public void run()
      {
         synchronized (SecurityManager.this)
         {
            SecurityCache sc = getCache();
   
            // establish initial security context
            SecurityContext context = sc.getInitialSecurityContext();
            logContext(sc, context);
            SecurityContextStack.setContext(new SecurityContextStack(context));
            assignContext(context);
   
            // hash the password
            byte[] hash = HashPassword.hashPassword(newPassword);
      
            // access the user account
            UserAccount ua = (UserAccount)sc.getAccountByOrd(ord);
            String uid = ua.getSubjectId();
            
            // change the password in the P2J directory
            String path = new String("/security/accounts/users/" + uid);
            
            // access the directory service
            DirectoryService ds = DirectoryService.getInstance();
            
            // is the directory service is unavailable?
            if (ds == null)
            {
               // done!
               return;
            }
            
            // try to bind
            if (!ds.bind())
            {
               // can't access the directory
               return;
            }
            
            // any failure from here out, triggers an unbind if needed
            Date date = new Date();
            DateValue cdate = new DateValue(date);
            TimeValue ctime = new TimeValue(date);
   
            try
            {
               if (!ds.openBatch(path))
                  throw new RuntimeException();
               
               // any failure from here out, triggers a rollback
               try
               {
                  // set password hash
                  if (!ds.setNodeByteArray(path, "password", hash))
                  {
                     if (!ds.addNodeByteArray(path, "password", hash))
                        throw new RuntimeException();
                  }
   
                  // set password modification date
                  if (!ds.setNodeDate(path, "pwsetdate", cdate))
                  {
                     if (!ds.addNodeDate(path, "pwsetdate", cdate))
                        throw new RuntimeException();
                  }
   
                  // set password modification time
                  if (!ds.setNodeTime(path, "pwsettime", ctime))
                  {
                     if (!ds.addNodeTime(path, "pwsettime", ctime))
                        throw new RuntimeException();
                  }
   
                  // commit the changes
                  if (!ds.closeBatch(true))
                     throw new RuntimeException();
   
                  // change the password in the cache
                  ua.setPassword(hash, cdate, ctime);
                  
                  result = true;
               }
               catch (Exception exc)
               {
                  // rollback the modification
                  ds.closeBatch(false);
               }
            }
            catch (Exception exc)
            {
               // just allow silent return
            }
            finally
            {
               // last use of the directory, end our session with it
               ds.unbind();
            }
         }
      }
   }

   /**
    * This class extends <code>ClassLoader</code> to enable the loading of
    * a class file stored as a byte array in memory.
    */
   private class HookClassLoader extends ClassLoader
   {
      /** byte array to load the class from */
      private byte[] data = null;

      /** 
       * Constructor for a single source.
       *
       * @param source
       *        array of bytes this class loader operates
       */
      private HookClassLoader(byte[] source)
      {
         data = source;
      }

      /** 
       * This method is called by the JVM when the class with the specified
       * name is referenced and needs to be loaded.  The primordial class
       * loader is first given a chance to load the class, in case <code>
       * name</code> refers to a system class.  If this load fails, an attempt
       * is then made to find the class in the cache maintained by this class
       * loader.  If the class is not found in the cache, a lookup is
       * performed to find the byte array containing the class file information
       * for the class identified by <code>name</code>, and this byte array is
       * used to define the class.
       *
       * @param   name
       *          Fully qualified name of the class to be loaded.
       * @param   resolve
       *          True if class should be resolved, else false.
       *
       * @return  The <code>Class</code> object identified by <code>name</code>.
       *
       * @exception  ClassNotFoundException
       *             If no class can be found for the specified name.
       */
      protected synchronized Class loadClass(String name, boolean resolve)
      throws ClassNotFoundException
      {
         Class c = null;
   
         try
         {
            c = findSystemClass(name);
         }
         catch (Exception exc)
         {
            try
            {
               c = defineClass(name, data, 0, data.length);
            }
            catch (Exception e)
            {
               exc.printStackTrace();
               throw new ClassNotFoundException(
                  "[" + e.getClass().getName() + "]  " + e.getMessage());
            }
         }
   
         if (resolve)
            resolveClass(c);
   
         return c;
      }
   }
}
